from nipype.pipeline import engine as pe
from nipype.interfaces import utility as niu
from nipype.interfaces.base import (
    traits, TraitedSpec, BaseInterfaceInputSpec,
    File, BaseInterface
)

def init_confound_correction_wf(cr_opts, name="confound_correction_wf"):
    # confound_wf_head_start
    """
    This workflow applies the RABIES confound correction pipeline to preprocessed EPI timeseries. The correction steps are 
    orchestrated in line with recommendations from human litterature:   
    #1 - Compute and apply frame censoring mask (from FD and/or DVARS thresholds)
    #2 - If --match_number_timepoints is selected, each scan is matched to the defined minimum_timepoint number of frames.
    #4 - Linear/Quadratic detrending of fMRI timeseries and nuisance regressors
    #4 - Apply ICA-AROMA.
    #5 - If frequency filtering and frame censoring are applied, simulate data in censored timepoints using the Lomb-Scargle periodogram, 
         as suggested in Power et al. (2014, Neuroimage), for both the fMRI timeseries and nuisance regressors prior to filtering.
    #6 - As recommended in Lindquist et al. (2019, Human brain mapping), make the nuisance regressors orthogonal
         to the temporal frequency filter.
    #7 - Apply highpass and/or lowpass filtering on the fMRI timeseries (with simulated timepoints).
    #8 - Re-apply the frame censoring mask onto filtered fMRI timeseries and nuisance regressors, taking out the
         simulated timepoints. Edge artefacts from frequency filtering can also be removed as recommended in Power et al. (2014, Neuroimage).
    #9 - Apply confound regression using the selected nuisance regressors.
    #10 - Scaling of timeseries variance.
    #11 - Apply Gaussian spatial smoothing.
    
    References:
        Power, J. D., Barnes, K. A., Snyder, A. Z., Schlaggar, B. L., & Petersen, S. E. (2012). Spurious but systematic 
            correlations in functional connectivity MRI networks arise from subject motion. Neuroimage, 59(3), 2142-2154.
        Power, J. D., Mitra, A., Laumann, T. O., Snyder, A. Z., Schlaggar, B. L., & Petersen, S. E. (2014). Methods to detect, 
            characterize, and remove motion artifact in resting state fMRI. Neuroimage, 84, 320-341.
        Lindquist, M. A., Geuter, S., Wager, T. D., & Caffo, B. S. (2019). Modular preprocessing pipelines can reintroduce 
            artifacts into fMRI data. Human brain mapping, 40(8), 2358-2376.

    Workflow:
        parameters
            cr_opts: command line interface parameters from confound_correction

        inputs
            bold_file: preprocessed EPI timeseries
            brain_mask: brain mask overlapping with EPI timeseries
            csf_mask: CSF mask overlapping with EPI timeseries
            motion_params_csv: CSV file with motion regressors
            FD_file: CSV file with the framewise displacement

        outputs
            cleaned_path: the cleaned EPI timeseries
            aroma_out: folder with outputs from ICA-AROMA
            VE_file: variance explained (R^2) from confound regression at each voxel
            STD_file: standard deviation on the cleaned EPI timeseries
            CR_STD_file: standard deviation on the confound timeseries modelled during confound regression
            random_CR_STD_file_path: variance fitted by random regressors during confound regression
            corrected_CR_STD_file_path: CR_STD_file after substracting the variance fitted by random
                regressors.
            frame_mask_file: CSV file which records which frame were censored
            CR_data_dict: dictionary object storing extra data computed during confound correction
    """
    # confound_wf_head_end

    workflow = pe.Workflow(name=name)
    inputnode = pe.Node(niu.IdentityInterface(fields=[
                        'bold_file', 'brain_mask', 'WM_mask', 'CSF_mask', 'vascular_mask', 'motion_params_csv', 'FD_file', 'raw_input_file']), name='inputnode')
    outputnode = pe.Node(niu.IdentityInterface(fields=[
                         'cleaned_path', 'aroma_out', 'VE_file', 'STD_file', 'CR_STD_file', 
                         'random_CR_STD_file_path', 'corrected_CR_STD_file_path', 'frame_mask_file', 'CR_data_dict']), name='outputnode')
    clean_image_node = pe.Node(CleanImage(cr_opts=cr_opts),
                           name='clean_image', mem_gb=3*cr_opts.scale_min_memory) # 3X memory as the timeseries is expanded into many arrays

    workflow.connect([
        (inputnode, clean_image_node, [
            ("bold_file", "bold_file"),
            ("brain_mask", "brain_mask_file"),
            ("WM_mask", "WM_mask_file"),
            ("CSF_mask", "CSF_mask_file"),
            ("vascular_mask", "vascular_mask_file"),
            ("raw_input_file", "raw_input_file"),
            ("motion_params_csv", "motion_params_csv"),
            ("FD_file", "FD_file"),
            ]),
        (clean_image_node, outputnode, [
            ("cleaned_path", "cleaned_path"),
            ("VE_file_path", "VE_file"),
            ("STD_file_path", "STD_file"),
            ("CR_STD_file_path", "CR_STD_file"),
            ("random_CR_STD_file_path", "random_CR_STD_file_path"),
            ("corrected_CR_STD_file_path", "corrected_CR_STD_file_path"),
            ("frame_mask_file", "frame_mask_file"),
            ("data_dict", "CR_data_dict"),
            ("aroma_out", "aroma_out"),
            ]),
        ])

    return workflow


class CleanImageInputSpec(BaseInterfaceInputSpec):
    raw_input_file = File(exists=True, mandatory=True,
                      desc="The raw EPI scan before preprocessing.")
    bold_file = File(exists=True, mandatory=True,
                      desc="Timeseries to denoise.")
    brain_mask_file = File(exists=True, mandatory=True,
                      desc="Brain mask.")
    WM_mask_file = traits.Any(mandatory=True,
                      desc="WM mask.")
    CSF_mask_file = traits.Any(mandatory=True,
                      desc="CSF mask.")
    vascular_mask_file = traits.Any(mandatory=True,
                      desc="vascular mask.")
    FD_file = File(exists=True, mandatory=True,
                      desc="CSV with framewise displacement.")
    motion_params_csv = File(exists=True, mandatory=True,
                      desc="CSV with motion parameters.")
    cr_opts = traits.Any(
        exists=True, mandatory=True, desc="Processing specs.")

class CleanImageOutputSpec(TraitedSpec):
    cleaned_path = File(exists=True, mandatory=True,
                      desc="Cleaned timeseries.")
    VE_file_path = File(exists=True, mandatory=True,
                      desc="Variance explained map from confound regression.")
    STD_file_path = File(exists=True, mandatory=True,
                      desc="Temporal standard deviation map after confound correction, prior to standardization.")
    CR_STD_file_path = File(exists=True, mandatory=True,
                      desc="Temporal standard deviation on predicted confound timeseries.")
    random_CR_STD_file_path = File(exists=False, mandatory=False,
                      desc="Temporal standard deviation on predicted confound timeseries from random regressors.")
    corrected_CR_STD_file_path = File(exists=False, mandatory=False,
                      desc="Same as CR_STD_file_path, but the variance explained by random regressors was substracted.")
    frame_mask_file = File(exists=True, mandatory=True,
                      desc="Frame mask from temporal censoring.")
    data_dict = traits.Any(
        desc="A dictionary with key outputs.")
    aroma_out = traits.Any(
        desc="Output directory from ICA-AROMA.")

class CleanImage(BaseInterface):
    '''
    Apply a flexible confound correction pipeline in line with recommendations from
    human litterature. 
    
    #1 - Compute and apply frame censoring mask (from FD and/or DVARS thresholds)
    #2 - If --match_number_timepoints is selected, each scan is matched to the defined minimum_timepoint number of frames.
    #4 - Linear/Quadratic detrending of fMRI timeseries and nuisance regressors
    #4 - Apply ICA-AROMA.
    #5 - If frequency filtering and frame censoring are applied, simulate data in censored timepoints using the Lomb-Scargle periodogram, 
         as suggested in Power et al. (2014, Neuroimage), for both the fMRI timeseries and nuisance regressors prior to filtering.
    #6 - As recommended in Lindquist et al. (2019, Human brain mapping), make the nuisance regressors orthogonal
         to the temporal frequency filter.
    #7 - Apply highpass and/or lowpass filtering on the fMRI timeseries (with simulated timepoints).
    #8 - Re-apply the frame censoring mask onto filtered fMRI timeseries and nuisance regressors, taking out the
         simulated timepoints. Edge artefacts from frequency filtering can also be removed as recommended in Power et al. (2014, Neuroimage).
    #9 - If selected, compute the WM/CSF/vascular signal or aCompCorr and add to list of regressors. This is computed post-AROMA and filtering to 
         minimize re-introduction of previously corrected signal fluctuations.
    #10 - Apply confound regression using the selected nuisance regressors.
    #11 - Scaling of timeseries.
    #12 - Apply Gaussian spatial smoothing.
    
    References:
        
        Power, J. D., Barnes, K. A., Snyder, A. Z., Schlaggar, B. L., & Petersen, S. E. (2012). 
        Spurious but systematic correlations in functional connectivity MRI networks arise from subject motion. Neuroimage, 59(3), 2142-2154.
        
        Power, J. D., Mitra, A., Laumann, T. O., Snyder, A. Z., Schlaggar, B. L., & Petersen, S. E. (2014). 
        Methods to detect, characterize, and remove motion artifact in resting state fMRI. Neuroimage, 84, 320-341.
        
        Lindquist, M. A., Geuter, S., Wager, T. D., & Caffo, B. S. (2019). 
        Modular preprocessing pipelines can reintroduce artifacts into fMRI data. Human brain mapping, 40(8), 2358-2376.
    '''

    input_spec = CleanImageInputSpec
    output_spec = CleanImageOutputSpec

    def _run_interface(self, runtime):
        import os
        import numpy as np
        import pandas as pd
        import SimpleITK as sitk

        ### set null returns in case the workflow is interrupted
        empty_img = sitk.GetImageFromArray(np.empty([1,1]))
        empty_file = os.path.abspath('empty.nii.gz')
        sitk.WriteImage(empty_img, empty_file)

        setattr(self, 'cleaned_path', empty_file)
        setattr(self, 'VE_file_path', empty_file)
        setattr(self, 'STD_file_path', empty_file)
        setattr(self, 'CR_STD_file_path', empty_file)
        setattr(self, 'random_CR_STD_file_path', empty_file)
        setattr(self, 'corrected_CR_STD_file_path', empty_file)
        setattr(self, 'frame_mask_file', empty_file)
        setattr(self, 'data_dict', empty_file)
        setattr(self, 'aroma_out', empty_file)
        ###

        cr_opts = self.inputs.cr_opts

        from nipype import logging
        nipype_log = logging.getLogger('nipype.workflow')

        with np.errstate(invalid='ignore', divide='ignore'):
            cleaning_out = clean_image(
                input_bold = self.inputs.bold_file,
                brain_mask = self.inputs.brain_mask_file,
                WM_mask = self.inputs.WM_mask_file,
                CSF_mask = self.inputs.CSF_mask_file,
                vascular_mask = self.inputs.vascular_mask_file,
                FD_csv = self.inputs.FD_file,
                motion_params_csv=self.inputs.motion_params_csv,
                timeseries_interval = cr_opts.timeseries_interval,
                FD_censoring=cr_opts.frame_censoring['FD_censoring'], 
                FD_threshold=cr_opts.frame_censoring['FD_threshold'], 
                DVARS_censoring=cr_opts.frame_censoring['DVARS_censoring'], 
                MSE_censoring=cr_opts.frame_censoring['MSE_censoring'], 
                minimum_timepoint=cr_opts.frame_censoring['minimum_timepoint'],
                TR=cr_opts.TR,
                detrending_order=cr_opts.detrending['order'], 
                detrending_time_interval=cr_opts.detrending['time_interval'], 
                apply_ica_aroma=cr_opts.ica_aroma['apply'], 
                ica_aroma_dim=cr_opts.ica_aroma['dim'], 
                ica_aroma_random_seed=cr_opts.ica_aroma['random_seed'],
                match_number_timepoints=cr_opts.match_number_timepoints, 
                highpass=cr_opts.highpass, 
                lowpass=cr_opts.lowpass, 
                edge_cutoff=cr_opts.edge_cutoff,
                nuisance_regressors = cr_opts.nuisance_regressors, 
                generate_CR_null=cr_opts.generate_CR_null,
                scale_variance_voxelwise=cr_opts.scale_variance_voxelwise,
                image_scaling=cr_opts.image_scaling,
                smoothing_filter=cr_opts.smoothing_filter,
                slicewise_correction_direction=cr_opts.slicewise_correction_direction,
                nipype_log=nipype_log,
                )

        if cleaning_out is None:
            return runtime
        else:
            [timeseries_img, 
            CR_data_dict, 
            VE_spatial_map, 
            STD_spatial_map, 
            CR_STD_spatial_map, 
            random_CR_STD_spatial_map, 
            corrected_CR_STD_spatial_map] = cleaning_out

        '''
        Generate all the output files
        '''
        cr_out = os.getcwd()
        import pathlib  # Better path manipulation
        filename_split = pathlib.Path(self.inputs.raw_input_file).name.rsplit(".nii")

        cleaned_path = cr_out+'/'+filename_split[0]+'_cleaned.nii.gz'
        sitk.WriteImage(timeseries_img, cleaned_path)
        VE_file_path = cr_out+'/'+filename_split[0]+'_VE_map.nii.gz'
        sitk.WriteImage(VE_spatial_map, VE_file_path)
        STD_file_path = cr_out+'/'+filename_split[0]+'_STD_map.nii.gz'
        sitk.WriteImage(STD_spatial_map, STD_file_path)
        CR_STD_file_path = cr_out+'/'+filename_split[0]+'_CR_STD_map.nii.gz'
        sitk.WriteImage(CR_STD_spatial_map, CR_STD_file_path)
        frame_mask_file = cr_out+'/'+filename_split[0]+'_frame_censoring_mask.csv'
        pd.DataFrame(CR_data_dict['frame_mask']).to_csv(frame_mask_file, index=False, header=['False = Masked Frames'])

        if cr_opts.generate_CR_null:
            random_CR_STD_file_path = cr_out+'/'+filename_split[0]+'_random_CR_STD_map.nii.gz'
            sitk.WriteImage(random_CR_STD_spatial_map, random_CR_STD_file_path)
            corrected_CR_STD_file_path = cr_out+'/'+filename_split[0]+'_corrected_CR_STD_map.nii.gz'
            sitk.WriteImage(corrected_CR_STD_spatial_map, corrected_CR_STD_file_path)
        else:
            random_CR_STD_file_path=''
            corrected_CR_STD_file_path=''

        CR_data_dict['cleaned_path'] = cleaned_path
        CR_data_dict['VE_file_path'] = VE_file_path
        CR_data_dict['STD_file_path'] = STD_file_path
        CR_data_dict['CR_STD_file_path'] = CR_STD_file_path
        CR_data_dict['random_CR_STD_file_path'] = random_CR_STD_file_path
        CR_data_dict['corrected_CR_STD_file_path'] = corrected_CR_STD_file_path
        CR_data_dict['frame_mask_file'] = frame_mask_file
                
        setattr(self, 'cleaned_path', cleaned_path)
        setattr(self, 'VE_file_path', VE_file_path)
        setattr(self, 'STD_file_path', STD_file_path)
        setattr(self, 'CR_STD_file_path', CR_STD_file_path)
        setattr(self, 'random_CR_STD_file_path', random_CR_STD_file_path)
        setattr(self, 'corrected_CR_STD_file_path', corrected_CR_STD_file_path)
        setattr(self, 'frame_mask_file', frame_mask_file)
        setattr(self, 'data_dict', CR_data_dict)
        if CR_data_dict['aroma_out'] is not None:
            setattr(self, 'aroma_out', CR_data_dict['aroma_out'])

        return runtime

    def _list_outputs(self):
        return {'cleaned_path': getattr(self, 'cleaned_path'),
                'VE_file_path': getattr(self, 'VE_file_path'),
                'STD_file_path': getattr(self, 'STD_file_path'),
                'CR_STD_file_path': getattr(self, 'CR_STD_file_path'),
                'random_CR_STD_file_path': getattr(self, 'random_CR_STD_file_path'),
                'corrected_CR_STD_file_path': getattr(self, 'corrected_CR_STD_file_path'),
                'frame_mask_file': getattr(self, 'frame_mask_file'),
                'data_dict': getattr(self, 'data_dict'),
                'aroma_out': getattr(self, 'aroma_out'),
                }
    

def clean_image(input_bold, brain_mask, FD_csv, motion_params_csv, # necessary input files
                WM_mask=None, CSF_mask=None, vascular_mask=None,
                timeseries_interval='0-end', FD_censoring=False, FD_threshold=0.05, 
                DVARS_censoring=False, MSE_censoring=False, minimum_timepoint=3, TR='auto',
                detrending_order=1, detrending_time_interval='0-end', 
                apply_ica_aroma=False, ica_aroma_dim=0, ica_aroma_random_seed=1,
                match_number_timepoints=False, highpass=None, lowpass=None, edge_cutoff=0,
                nuisance_regressors = [], generate_CR_null=False,
                scale_variance_voxelwise=False,image_scaling='grand_mean_scaling',
                smoothing_filter=None,
                slicewise_correction_direction = 'Off',
                nipype_log=None,
                ):
    """
    Core function for carying confound correction of an image using the set of RABIES tools.

    Parameters
    ----------
    input_bold : filepath or sitk.Image
        The input 4D timeseries image to clean.

    brain_mask : filepath or sitk.Image
        A brain mask in the same space as input_bold

    FD_csv : filepath or pd.DataFrame
        The framewise displacement CSV generated by RABIES associated with input_bold.

    motion_params_csv : filepath or pd.DataFrame
        The motion parameters CSV generated by RABIES associated with input_bold.
                
    WM_mask : filepath or sitk.Image, default=None
        A WM mask in the same space as input_bold

    CSF_mask : filepath or sitk.Image, default=None
        A CSV mask in the same space as input_bold

    vascular_mask : filepath or sitk.Image, default=None
        A vascular mask in the same space as input_bold
                
    timeseries_interval : str, default='0-end'
        The chunk of frames to clean controlled by --timeseries_interval.

    FD_censoring : bool, default=False
        Whether to apply FD censoring.
        
    FD_threshold : float, default=0.05
        The FD censoring threshold.

    DVARS_censoring : bool, default=False
        Whether to apply DVARS censoring.

    MSE_censoring : bool, default=False
        Whether to apply MSE censoring.

    minimum_timepoint : int, default=3
        Minimum number of frames left post-cleaning, otherwise returns None.

    TR : str, default='auto'
        The repetition time in seconds. If 'auto', the TR is read from the Nifti header.

    detrending_order : int, default=1
        The polynomial order for detrending.

    detrending_time_interval : str, default='0-end'
        The time interval over which the trends are computed for detrending. The trends
        are always removed across the entire timeseries, even if they were estimated from
        a subset of frames.
        This follows the same syntax as timeseries_interval.

    apply_ica_aroma : bool, default=False
        Whether to apply ICA-AROMA.

    ica_aroma_dim : int, default=0
        ICA dimensionality for AROMA. If 0, uses MELODIC's automated estimator.

    ica_aroma_random_seed : int, default=1
        Sets a random seed for AROMA to ensure deterministic behavior.

    match_number_timepoints : bool, default=False
        Whether to enfore a consistent final number of frames across all images that
        corresponds to minimum_timepoint, where extra frames are randomly removed.

    highpass : float, default=None
        Frequency cutoff for highpass filter.

    lowpass : float, default=None
        Frequency cutoff for lowpass filter.

    edge_cutoff : int, default=0
        Number of seconds of data to remove at each edge after frequency filtering to 
        remove edge artefacts.

    nuisance_regressors : list, default=[]
        List of nuisance regressors. Possible options are: "WM_signal", "CSF_signal", 
        "vascular_signal", "global_signal", "aCompCor_percent", "aCompCor_5", "mot_6", 
        "mot_24"

    generate_CR_null : bool, default=False
        Whether to estimate overfitting from confound regression by generating phase 
        randomised regressors, and comparing the variance explained by the real 
        regressors VS the randomized regressors.

    scale_variance_voxelwise : bool, default=False
        If scaling was not applied voxelwise with voxelwise_standardization or 
        voxelwise_mean, this option standardize the variance at each voxel to be equal, 
        while preserving to total variance of the 4D timeseries (i.e. voxels have same 
        variance, but not unit variance).

    image_scaling : str, default='grand_mean_scaling'
        Method for image scaling, select among: "None", "global_variance", 
        "voxelwise_standardization", "grand_mean_scaling", "voxelwise_mean"

    smoothing_filter : float, default=None
        Gaussian smoothing filter to apply.

    slicewise_correction_direction : str, default='Off'
        By inputing a slice direction with this parameters, the computation of nuisance 
        signals, nuisance regression, and smoothing, are all conducted on each 2D slice 
        independently along the provided slice direction. This strategy can be useful in 
        the presence of artefacts (e.g. intensity spikes) that are not shared across 
        slices due to time delays between slice acquisition, in which case a slicewise 
        nuisance regression may perform better. Otherwise, in the case of images with 
        spatial gaps between slices, smoothing should be applied slicewise using this 
        parameter. To apply slicewise correction, provide as input one of the RAS axial 
        direction for slice selection, i.e. this parameter takes as input one of: 'RL', 
        'AP' or 'SI'. For accurate slice selection, the axial directions of the input 
        nifti images must be properly encoded. If no direction is provided, standard 
        whole-brain correction is conducted.

    nipype_log : nipype.logging.getLogger('nipype.workflow'), default=None
        Provide a nipype log for printing statements.

    Returns
    -------

    timeseries_img : sitk.Image
        Output cleaned timeseries image.

    CR_data_dict : dict
        Output cleaned timeseries image.

    VE_spatial_map : sitk.Image
        R2 of nuisance regression voxelwise.

    STD_spatial_map : sitk.Image
        Voxelwise standard deviation of cleaned timeseries.

    CR_STD_spatial_map : sitk.Image
        Voxelwise standard deviation of fitted nuisance timeseries.

    random_CR_STD_spatial_map : sitk.Image
        Voxelwise standard deviation of fitted randomized regressors with 
        generate_CR_null=True.

    corrected_CR_STD_spatial_map : sitk.Image
        Voxelwise standard deviation of fitted nuisance timeseries, after correction 
        by the randomized regressors with generate_CR_null=True.
    
    Returns if internal error.
    -------
    None : if an error is detected in the workflow, the function only returns None.

    """    
    import os
    import numpy as np
    import pandas as pd
    import SimpleITK as sitk
    from rabies.utils import recover_3D,recover_4D
    from . import utils as cr_utils

    '''
    The function can take as input either an adequately pre-loaded python object or a file path
    '''
    def read_input(input):
        if (input is None) or isinstance(input, sitk.Image) or isinstance(input, pd.DataFrame):
            return input
        elif os.path.isfile(input):
            if '.nii' in input:
                return sitk.ReadImage(input, sitk.sitkFloat32)
            elif '.csv' in input:
                return pd.read_csv(input)
            else:
                raise ValueError(f"The input {input} was neither a .nii or .csv file.")
        else:
            raise ValueError(f"The input {input} was neither a file nor a recognized python object.")

    [input_bold, 
     FD_csv, 
     motion_params_df, 
     brain_mask, 
     WM_mask, 
     CSF_mask, 
     vascular_mask] = [read_input(input) for input in [
         input_bold, 
         FD_csv, 
         motion_params_csv, 
         brain_mask, 
         WM_mask, 
         CSF_mask, 
         vascular_mask,
         ]]

    volume_idx = sitk.GetArrayFromImage(brain_mask).astype(bool)
     # save the header and original array size, since the image will be removed to space memory below
    from rabies.utils import get_geometry_header
    header_geo = get_geometry_header(input_bold)
    orig_4d_size = input_bold.GetSize() # copy the original array size

    if TR=='auto':
        TR = float(header_geo.GetSpacing()[3])
    else:
        TR = float(TR)

    # save specifically the 6 rigid parameters for AROMA
    motion6_regressors_array = cr_utils.select_motion_regressors(['mot_6'],motion_params_df)
    if len(nuisance_regressors)==0:
        motion_regressors_array = motion6_regressors_array
    else:
        motion_regressors_array = cr_utils.select_motion_regressors(nuisance_regressors,motion_params_df)

    FD_trace = FD_csv.get('MeanFD')

    time_range = cr_utils.prep_timeseries_interval(timeseries_interval, num_frames=orig_4d_size[3])

    timeseries_4d_arr = sitk.GetArrayFromImage(input_bold)
    timeseries_4d_arr = timeseries_4d_arr[time_range]
    del input_bold
    mse_trace = cr_utils.MSE_relative_to_average(timeseries_4d_arr)
    timeseries = timeseries_4d_arr[:,volume_idx] # read directly as a 2D array
    del timeseries_4d_arr # free memory
    motion_regressors_array = motion_regressors_array[time_range, :]
    motion6_regressors_array = motion6_regressors_array[time_range, :]
    FD_trace = FD_trace[time_range]

    '''
    #1 - Compute and apply frame censoring mask (from FD and/or DVARS thresholds)
    '''

    # compute the DVARS before denoising
    DVARS_trace = cr_utils.get_DVARS(timeseries)

    frame_mask = cr_utils.temporal_censoring(FD_trace, 
            FD_censoring, FD_threshold, DVARS_trace, DVARS_censoring, 
            mse_trace, MSE_censoring, minimum_timepoint)
    if frame_mask is None:
        return None

    '''
    #2 - If --match_number_timepoints is selected, each scan is matched to the defined minimum_timepoint number of frames.
    '''
    if match_number_timepoints:
        if (highpass is not None) or (lowpass is not None):
            # if frequency filtering is applied, avoid selecting timepoints that would be removed with --edge_cutoff
            num_cut = int(edge_cutoff/TR)
            if not num_cut==0:
                frame_mask[:num_cut]=0
                frame_mask[-num_cut:]=0

                if frame_mask.sum()<int(minimum_timepoint):
                    if nipype_log:
                        nipype_log.warning(f"CONFOUND CORRECTION LEFT LESS THAN {str(minimum_timepoint)} VOLUMES. THIS SCAN WILL BE REMOVED FROM FURTHER PROCESSING.")
                    return None

        # randomly shuffle indices that haven't been censored, then remove an extra subset above --minimum_timepoint
        num_timepoints = len(frame_mask)
        time_idx=np.array(range(num_timepoints))
        perm = np.random.permutation(time_idx[frame_mask])
        # selecting the subset of extra timepoints, and censoring them
        subset_idx = perm[minimum_timepoint:]
        frame_mask[subset_idx]=0
        # keep track of the original number of timepoints for tDOF estimation, to evaluate latter if the correction was succesful
        number_extra_timepoints = len(subset_idx)
    else:
        number_extra_timepoints = 0

    timeseries = timeseries[frame_mask]
    motion_regressors_array = motion_regressors_array[frame_mask]

    '''
    #3 - Linear/Quadratic detrending of fMRI timeseries and nuisance regressors
    '''
    timeseries, voxelwise_intercept = cr_utils.remove_trend(timeseries, frame_mask, order = detrending_order, time_interval = detrending_time_interval)
    motion_regressors_array, _ = cr_utils.remove_trend(motion_regressors_array, frame_mask, order = detrending_order, time_interval = detrending_time_interval)

    '''
    #4 - Apply ICA-AROMA.
    '''
    if apply_ica_aroma:
        # write intermediary output files for timeseries and 6 rigid body parameters
        inFile = os.path.abspath('aroma_input_timeseries.nii.gz')
        sitk.WriteImage(recover_4D(brain_mask, timeseries, header_geo), inFile)
        brain_mask_file = os.path.abspath("aroma_brain_mask.nii.gz")
        sitk.WriteImage(brain_mask, brain_mask_file)
        CSF_mask_file = os.path.abspath("aroma_CSF_mask.nii.gz")
        sitk.WriteImage(CSF_mask, CSF_mask_file)

        motion6_regressors_array=motion6_regressors_array[frame_mask,:]
        motion6_regressors_array, _ = cr_utils.remove_trend(motion6_regressors_array, frame_mask, order = detrending_order, time_interval = detrending_time_interval) # apply detrending to the confounds too
        df = pd.DataFrame(motion6_regressors_array)
        df.columns = ['mov1', 'mov2', 'mov3', 'rot1', 'rot2', 'rot3']
        mc_file = os.path.abspath('aroma_motion_params.csv')
        df.to_csv(mc_file)
        aroma_out = os.path.abspath('aroma_out/')
        cleaned_file, aroma_out = cr_utils.exec_ICA_AROMA(inFile, mc_file, brain_mask_file, CSF_mask_file, aroma_out, TR, ica_aroma_dim, random_seed=ica_aroma_random_seed)
        timeseries = sitk.GetArrayFromImage(sitk.ReadImage(cleaned_file, sitk.sitkFloat32))[:,volume_idx] # read directly as a 2D array
    else:
        aroma_out = None

    if (highpass is not None) or (lowpass is not None):
        '''
        #5 - If frequency filtering and frame censoring are applied, simulate data in censored timepoints using the Lomb-Scargle periodogram, 
            as suggested in Power et al. (2014, Neuroimage), for both the fMRI timeseries and nuisance regressors prior to filtering.
        '''
        timeseries = cr_utils.lombscargle_fill(x=timeseries,time_step=TR,time_mask=frame_mask)
        motion_regressors_array = cr_utils.lombscargle_fill(x=motion_regressors_array,time_step=TR,time_mask=frame_mask)
        ### arrays are now interpolated

        '''
        #6 - As recommended in Lindquist et al. (2019, Human brain mapping), make the nuisance regressors orthogonal
            to the temporal filter.
        '''
        motion_regressors_array = cr_utils.butterworth(motion_regressors_array, TR=TR,
                                high_pass=highpass, low_pass=lowpass)

        '''
        #7 - Apply highpass and/or lowpass filtering on the fMRI timeseries (with simulated timepoints).
        '''

        timeseries = cr_utils.butterworth(timeseries, TR=TR,
                                high_pass=highpass, low_pass=lowpass)

        # correct for edge effects of the filters
        num_cut = int(edge_cutoff/TR)
        if len(frame_mask)<2*num_cut:
            raise ValueError(f"The timeseries are too short to remove {edge_cutoff}sec of data at each edge.")

        if not num_cut==0:
            frame_mask[:num_cut]=0
            frame_mask[-num_cut:]=0


        '''
        #8 - Re-apply the frame censoring mask onto filtered fMRI timeseries and nuisance regressors, taking out the
            simulated timepoints. Edge artefacts from frequency filtering can also be removed as recommended in Power et al. (2014, Neuroimage).
        '''
        # re-apply the masks to take out simulated data points, and take off the edges
        timeseries = timeseries[frame_mask]
        motion_regressors_array = motion_regressors_array[frame_mask]
    
    if frame_mask.sum()<int(minimum_timepoint):
        if nipype_log:
            nipype_log.warning(f"CONFOUND CORRECTION LEFT LESS THAN {str(minimum_timepoint)} VOLUMES. THIS SCAN WILL BE REMOVED FROM FURTHER PROCESSING.")
        return None

    '''
    #9 - If selected, compute the WM/CSF/vascular signal or aCompCorr and add to list of regressors. This is computed post-AROMA and filtering to 
        minimize re-introduction of previously corrected signal fluctuations.
    #10 - Apply confound regression using the selected nuisance regressors.
    '''
    regress_out = cr_utils.nuisance_regression(
        timeseries, motion_regressors_array, TR, frame_mask, orig_4d_size, brain_mask, WM_mask, CSF_mask, vascular_mask, nuisance_regressors=nuisance_regressors, 
        slicewise_correction_direction=slicewise_correction_direction, generate_CR_null=generate_CR_null, nipype_log=nipype_log)
    if regress_out is None:
        return None
    timeseries, predicted, predicted_random, num_regressors, VE_temporal, VE_spatial, VE_total_ratio, cleaned_slice_l = regress_out

    '''
    #11 - Scaling of timeseries.
    '''
    if scale_variance_voxelwise: # homogenize the variability distribution while preserving the same total variance
        if image_scaling=='voxelwise_standardization' or image_scaling=='voxelwise_mean':
            raise ValueError(f"Can't select --scale_variance_voxelwise with --image_scaling {image_scaling}.")
        else:
            temporal_std = timeseries.std(axis=0)
            total_std = timeseries.std()

            # homogenize variance across voxels
            timeseries /= temporal_std
            nan_voxels = np.isnan(timeseries).sum(axis=0)>1
            timeseries[:,nan_voxels] = 0
            scaled_total_std = timeseries.std()
            # rescale to preserve total variance
            timeseries = (timeseries/scaled_total_std)*total_std

            # the same scaling estimated from BOLD is applied to CR estimates to preserve accurate ratios of variance explained per voxel
            scaled_list = []
            for scaled_timeseries in [predicted,predicted_random]:
                scaled_timeseries /= temporal_std
                nan_voxels = np.isnan(scaled_timeseries).sum(axis=0)>1
                scaled_timeseries[:,nan_voxels] = 0
                scaled_timeseries = (scaled_timeseries/scaled_total_std)*total_std
                scaled_list.append(scaled_timeseries)
            del scaled_timeseries
            [predicted,predicted_random] = scaled_list


    if image_scaling=='global_variance':
        scaling_factor = timeseries.std()
    elif image_scaling=='grand_mean_scaling':
        scaling_factor = voxelwise_intercept.mean()/100 # we scale BOLD in % fluctuations, hence dividing by 100
    elif image_scaling=='voxelwise_standardization':
        # each voxel is scaled according to its STD
        scaling_factor = timeseries.std(axis=0) 
    elif image_scaling=='voxelwise_mean':
        scaling_factor = voxelwise_intercept/100 # we scale BOLD in % fluctuations, hence dividing by 100
    else:
        scaling_factor = 1

    # scale fMRI timeseries, as well as the CR prediction to keep consistent scaling
    scaled_list = []
    for scaled_timeseries in [timeseries,predicted,predicted_random]:
        scaled_timeseries /= scaling_factor
        nan_voxels = np.isnan(scaled_timeseries).sum(axis=0)>1
        scaled_timeseries[:,nan_voxels] = 0
        scaled_list.append(scaled_timeseries)
    del scaled_timeseries
    [timeseries,predicted,predicted_random] = scaled_list

    # after variance scaling, compute the variability estimates
    temporal_std = timeseries.std(axis=0)
    predicted_std = predicted.std(axis=0)
    predicted_time = np.sqrt((predicted.T**2).mean(axis=0))
    predicted_global_std = predicted.std()

    if generate_CR_null:
        predicted_random_std = predicted_random.std(axis=0)

        # here we correct the previous STD estimates by substrating the variance explained by that of the overfitting with random regressors
        var_dif = predicted.var(axis=0) - predicted_random.var(axis=0)
        var_dif[var_dif<0] = 0 # when there's more variance explained in random regressors, set variance explained to 0
        corrected_predicted_std = np.sqrt(var_dif)
        del predicted_random
    del predicted

    # apply the frame mask to FD trace/DVARS to prepare outputs from the function
    DVARS_trace = DVARS_trace[frame_mask]
    FD_trace = FD_trace[frame_mask]
    mse_trace = mse_trace[frame_mask]

    # calculate temporal degrees of freedom left after confound correction
    num_timepoints = frame_mask.sum()
    if apply_ica_aroma:
        aroma_rm = (pd.read_csv(f'{aroma_out}/classification_overview.txt', sep='\t')['Motion/noise']).sum()
    else:
        aroma_rm = 0
    tDOF = num_timepoints - (aroma_rm+num_regressors) + number_extra_timepoints

    # compute DVARS correlation with FD post-correction
    DVARS_cleaned_trace = cr_utils.get_DVARS(timeseries)
    FD_DVARS_corr = np.corrcoef(DVARS_cleaned_trace[1:], FD_trace[1:])[0,1] # skip first index which is 0 for DVARS

    # smoothing takes an SITK image
    timeseries_img = recover_4D(brain_mask, timeseries, header_geo)
    del timeseries

    if smoothing_filter is not None:
        '''
        #12 - Apply Gaussian spatial smoothing.
        '''
        if not slicewise_correction_direction=='Off':
            # smooth 1 slice at a time
            for slice in cleaned_slice_l:
                # here the axis direction is in proper order since we are indexing a SITK image
                if slicewise_correction_direction=='RL':
                    timeseries_img[slice:slice+1,:,:,:] = cr_utils.smooth_image(
                        timeseries_img[slice:slice+1,:,:,:], smoothing_filter, brain_mask[slice:slice+1,:,:])
                elif slicewise_correction_direction=='AP':
                    timeseries_img[:,slice:slice+1,:,:] = cr_utils.smooth_image(
                        timeseries_img[:,slice:slice+1,:,:], smoothing_filter, brain_mask[:,slice:slice+1,:])
                elif slicewise_correction_direction=='SI':
                    timeseries_img[:,:,slice:slice+1,:] = cr_utils.smooth_image(
                        timeseries_img[:,:,slice:slice+1,:], smoothing_filter, brain_mask[:,:,slice:slice+1])
                else:
                    raise ValueError(f"slicewise_correction_direction must be one of 'RL', 'AP' or 'SI', got {slicewise_correction_direction} instead.")
        else:
            timeseries_img = cr_utils.smooth_image(
                timeseries_img, smoothing_filter, brain_mask)

    # save output files
    VE_spatial_map = recover_3D(brain_mask, VE_spatial)
    STD_spatial_map = recover_3D(brain_mask, temporal_std)
    CR_STD_spatial_map = recover_3D(brain_mask, predicted_std)
    del VE_spatial, temporal_std, predicted_std
    if generate_CR_null:
        random_CR_STD_spatial_map = recover_3D(brain_mask, predicted_random_std)
        corrected_CR_STD_spatial_map = recover_3D(brain_mask, corrected_predicted_std)
        del predicted_random_std, corrected_predicted_std
    else:
        random_CR_STD_spatial_map = None
        corrected_CR_STD_spatial_map = None

    CR_data_dict = {
        'TR':TR, 'FD_trace':FD_trace, 'DVARS_pre_correction':DVARS_trace, 'DVARS_post_correction':DVARS_cleaned_trace, 'FD_DVARS_corr':FD_DVARS_corr, 'mse_trace':mse_trace, 'time_range':time_range, 'frame_mask':frame_mask, 'VE_temporal':VE_temporal, 
        'motion_params_df':motion_params_df, 'predicted_time':predicted_time, 'tDOF':tDOF, 'CR_global_std':predicted_global_std, 
        'VE_total_ratio':VE_total_ratio, 'voxelwise_mean':voxelwise_intercept, 'aroma_out':aroma_out,
        }
    return timeseries_img, CR_data_dict, VE_spatial_map, STD_spatial_map, CR_STD_spatial_map, random_CR_STD_spatial_map, corrected_CR_STD_spatial_map



import os
import pathlib
from nipype.interfaces.base import (
    traits, TraitedSpec, BaseInterfaceInputSpec,
    File, BaseInterface
)
from nipype.pipeline import engine as pe
from nipype.interfaces import utility as niu

def init_bold_hmc_wf(opts, name='bold_hmc_wf'):
    # hmc_wf_head_start
    """
    This workflow estimates motion during fMRI acquisition. To do so, each EPI frame is registered to a volumetric 
    target reference image with a rigid registration. 
    This results in the measurement of 3 Euler angles in radians and 3 translations in mm (from ITK's 
    Euler3DTransform https://itk.org/Doxygen/html/classitk_1_1Euler3DTransform.html) at each time frame, which are 
    then stored into an output CSV file.

    Command line interface parameters:
        --HMC_level {1,2,3,4}
                                Select a level of stringence for the framewise registration (the higher the level, the 
                                more stringent it is).
                                (default: 2)

    Workflow:
        parameters
            opts: command line interface parameters

        inputs
            bold_file: Nifti file with EPI timeseries to realign
            ref_image: the 3D image target for realignment

        outputs
            motcorr_params: CSV file which contains all translation and rotation parameters
    """
    # hmc_wf_head_end

    workflow = pe.Workflow(name=name)
    inputnode = pe.Node(niu.IdentityInterface(fields=['bold_file', 'ref_image']),
                        name='inputnode')
    outputnode = pe.Node(
        niu.IdentityInterface(
            fields=['motcorr_params', 'hmc_qc_figure', 'hmc_qc_csv', 'hmc_qc_video']),
        name='outputnode')

    # Head motion correction (hmc)
    n_procs=int(os.environ['RABIES_ITK_NUM_THREADS'])
    motion_estimation = pe.Node(sitkMotionCorr(level=opts.HMC_level,rabies_data_type=opts.data_type),
                         name='motion_correction', mem_gb=1.1*opts.scale_min_memory, n_procs=n_procs)
    motion_estimation.plugin_args = {
        'qsub_args': f'-pe smp {str(3*opts.min_proc)}', 'overwrite': True}
    
    workflow.connect([
        (inputnode, motion_estimation, [('ref_image', 'ref_file'),
                                        ('bold_file', 'in_file')]),
        (motion_estimation, outputnode, [
         ('csv_params', 'motcorr_params')]),
    ])

    if opts.hmc_qc_report['apply']:
         # multiply mem_gb 20X because there is massive RAM overload with this node
        hmc_qc_node = pe.Node(HMC_QC(fps=opts.hmc_qc_report['fps'], figure_format=opts.figure_format),
                            name='hmc_qc_node', mem_gb=1.1*opts.scale_min_memory*20, n_procs=n_procs)

        workflow.connect([
            (inputnode, hmc_qc_node, [('ref_image', 'ref_file'),
                                            ('bold_file', 'in_file')]),
            (motion_estimation, hmc_qc_node, [
            ('csv_params', 'csv_params')]),
            (hmc_qc_node, outputnode, [
                ('out_figure', 'hmc_qc_figure'),
                ('out_csv', 'hmc_qc_csv'),
                ('video_file', 'hmc_qc_video'),
            ]),
        ])

    return workflow


class sitkMotionCorrInputSpec(BaseInterfaceInputSpec):
    in_file = File(exists=True, mandatory=True, desc='input BOLD time series')
    ref_file = File(exists=True, mandatory=True,
                    desc='ref file to realignment time series')
    level = traits.Int(desc="Select a level from 1 to 4, where each level is more stringent registration.")
    rabies_data_type = traits.Int(mandatory=True,
                                  desc="Integer specifying SimpleITK data type.")


class sitkMotionCorrOutputSpec(TraitedSpec):
    csv_params = File(
        exists=True, desc="csv files with the 6-parameters rigid body transformations")


class sitkMotionCorr(BaseInterface):
    """
    This interface performs motion realignment using simpleitk_timeseries_motion_correction. It takes a reference volume to which
    EPI volumes from the input 4D file are realigned based on a Rigid registration.
    """

    input_spec = sitkMotionCorrInputSpec
    output_spec = sitkMotionCorrOutputSpec

    def _run_interface(self, runtime):

        n_procs = int(os.environ['RABIES_ITK_NUM_THREADS']) if "RABIES_ITK_NUM_THREADS" in os.environ else os.cpu_count() # default to number of CPUs
        import SimpleITK as sitk
        from simpleitk_timeseries_motion_correction.motion import framewise_register_pair, write_transforms_to_csv
        moving = self.inputs.in_file
        ref_file = self.inputs.ref_file
        level = self.inputs.level
        filename_split = pathlib.Path(moving).name.rsplit(".nii")
        output_prefix = os.path.abspath(
            f'{filename_split[0]}_')

        transforms = framewise_register_pair(
            moving, 
            ref_file, 
            level=level, 
            interpolation=sitk.sitkBSpline5, 
            max_workers=n_procs)
        
        csv_param_file = output_prefix + f"moco.csv"
        write_transforms_to_csv(transforms, csv_param_file)
        setattr(self, 'csv_params', csv_param_file)

        return runtime

    def _list_outputs(self):
        return {'csv_params': getattr(self, 'csv_params')}


class EstimateMotionParamsInputSpec(BaseInterfaceInputSpec):
    motcorr_params = File(exists=True, mandatory=True,
                       desc="CSV file with the 6 rigid body parameters")
    boldspace_bold = File(exists=True, mandatory=True,
                      desc="BOLD image in its original space.")
    boldspace_brain_mask = File(exists=True, mandatory=True,
                      desc="Brain mask in the BOLD space.")


class EstimateMotionParamsOutputSpec(TraitedSpec):
    motion_params_csv = traits.File(desc="CSV file of motion parameters")
    FD_csv = traits.File(desc="CSV file with global framewise displacement.")


class EstimateMotionParams(BaseInterface):
    # motion_param_head_start
    """
    This interface generates estimations of framewise displacement, together with
    the expansion of the 6 motion parameters to include derivatives and squared parameters (Friston 24).
    Framewise displacement are computed within as follows:
        1. For each timepoint, the 3 Euler rotations and translations are converted to an affine matrix
        2. For each voxel within a brain mask representing the referential space post-motion realignment,
           the inverse transform is applied to generate a point pre-motion realignment.
        3. For framewise displacement, the distance is measured between the pre-correction points generated 
           from the current and the previous timeframe. Distance is measured in mm with the Euclidean distance.
    """
    # motion_param_head_end

    input_spec = EstimateMotionParamsInputSpec
    output_spec = EstimateMotionParamsOutputSpec

    def _run_interface(self, runtime):
        import numpy as np
        import os
        import pathlib  # Better path manipulation
        from simpleitk_timeseries_motion_correction.framewise_displacement import calculate_framewise_displacement

        filename_split = pathlib.Path(self.inputs.boldspace_bold).name.rsplit(".nii")

        FD_csv = os.path.abspath(f"{filename_split[0]}_FD_file.csv")
        calculate_framewise_displacement(mask_file=self.inputs.boldspace_brain_mask, csv_file=self.inputs.motcorr_params, output_csv=FD_csv, verbose=False)

        motion_24,motion_24_header = motion_24_params(self.inputs.motcorr_params)
        # write into a .csv
        import pandas as pd
        df = pd.DataFrame(motion_24)
        df.columns = motion_24_header
        motion_params_csv = os.path.abspath(f"{filename_split[0]}_motion_params.csv")
        df.to_csv(motion_params_csv)

        setattr(self, 'FD_csv', FD_csv)
        setattr(self, 'motion_params_csv', motion_params_csv)

        return runtime

    def _list_outputs(self):
        return {'motion_params_csv': getattr(self, 'motion_params_csv'),
                'FD_csv': getattr(self, 'FD_csv')}


def motion_24_params(movpar_csv):
    '''
    motioncorr_24params: 6 head motion parameters, their temporal derivative, and the 12 corresponding squared items (Friston et al. 1996, Magn. Reson. Med.)
    '''
    import numpy as np
    import pandas as pd
    MOCO_df = pd.read_csv(movpar_csv)
    rigid_params = np.array([MOCO_df[param].values for param in ['EulerX','EulerY','EulerZ','TransX','TransY','TransZ']]).T

    rotations = rigid_params[:,:3] # rotations are listed first
    translations = rigid_params[:,3:] # translations are last

    movpar = np.zeros([np.size(rigid_params, 0), 24])
    movpar[:, :3] = translations # add rotations first
    movpar[:, 3:6] = rotations # rotations second
    for i in range(6):
        # Compute temporal derivative as difference between two neighboring points
        movpar[0, 6+i] = 0
        movpar[1:, 6+i] = movpar[1:, i]-movpar[:-1, i]
        # add the squared coefficients
        movpar[:, 12+i] = movpar[:, i]**2
        movpar[:, 18+i] = movpar[:, 6+i]**2

    motion_24_header = ['mov1', 'mov2', 'mov3', 'rot1', 'rot2', 'rot3', 'mov1_der', 'mov2_der', 'mov3_der', 'rot1_der', 'rot2_der', 'rot3_der',
                    'mov1^2', 'mov2^2', 'mov3^2', 'rot1^2', 'rot2^2', 'rot3^2', 'mov1_der^2', 'mov2_der^2', 'mov3_der^2', 'rot1_der^2', 'rot2_der^2', 'rot3_der^2']

    return movpar,motion_24_header


class HMC_QCInputSpec(BaseInterfaceInputSpec):
    in_file = File(exists=True, mandatory=True, desc='input BOLD time series')
    ref_file = File(exists=True, mandatory=True,
                    desc='ref file to realignment time series')
    csv_params = File(
        exists=True, desc="csv files with the 6-parameters rigid body transformations")
    fps = traits.Int(default=10,
        desc="Frames per second for the video.")
    figure_format = traits.Str(default='png',
        mandatory=True, exists=True, desc="png or svg")


class HMC_QCOutputSpec(TraitedSpec):
    out_figure = File(exists=True, desc="Output figure for QC visualization.")
    out_csv = File(exists=True, desc="Output CSV saving certain derivatives.")
    video_file = File(exists=True, desc="Output video showing timeseries pre/post correction.")


class HMC_QC(BaseInterface):
    """
    This interface pairs with the outputs from the head motion correction interface to generative a qualitative report
    that evaluate the performance of the correction.
    """

    input_spec = HMC_QCInputSpec
    output_spec = HMC_QCOutputSpec

    def _run_interface(self, runtime):
        import pandas as pd
        from simpleitk_timeseries_motion_correction.create_animation import main
        filename = pathlib.Path(self.inputs.in_file).name.rsplit(".nii")[0]
        figure_path = os.path.abspath(f'{filename}_HMC_QC.{self.inputs.figure_format}')
        csv_path = os.path.abspath(f'{filename}_derivatives.csv')

        n_procs = int(os.environ['RABIES_ITK_NUM_THREADS']) if "RABIES_ITK_NUM_THREADS" in os.environ else os.cpu_count() # default to number of CPUs

        print('Calculating HMC derivatives for QC visualization...')
        img_preHMC, img_postHMC, derivatives_dict = HMC_derivatives(self.inputs.in_file, self.inputs.ref_file, self.inputs.csv_params, get_R2=False, n_procs=n_procs)

        print('Creating video pre/post HMC...')
        # write .webp file
        video_file = os.path.abspath(f'{filename}_HMC.webp')
        main(input_img=img_preHMC, 
             output_file=video_file, 
             additional_input_imgs=[img_postHMC], 
             labels=['Before correction', 'After correction'], 
             scale=2.0, 
             fps=self.inputs.fps)
        del img_preHMC, img_postHMC

        # save some outputs to .csv
        key_l = ['D_Sc_preHMC', 'D_Sc_postHMC', 'mse_preHMC', 'mse_postHMC']
        pd.DataFrame(np.array([derivatives_dict[key].flatten() for key in key_l]).T, columns=key_l).to_csv(csv_path)

        print('Creating QC figure...')
        fig = plot_motion_QC(derivatives_dict, self.inputs.ref_file, plot_R2=False)
        fig.savefig(figure_path, bbox_inches='tight')

        setattr(self, 'out_figure', figure_path)
        setattr(self, 'out_csv', csv_path)
        setattr(self, 'video_file', video_file)

        return runtime

    def _list_outputs(self):
        return {'out_figure': getattr(self, 'out_figure'),
                'out_csv': getattr(self, 'out_csv'),
                'video_file': getattr(self, 'video_file'),
                }


#############FUNCTIONS FOR HMC QC
import numpy as np
import SimpleITK as sitk
from rabies.utils import copyInfo_3DImage

def cosine_similarity(X,Y): 
    X_ = X.copy()
    X_ /= np.sqrt((X_ ** 2).sum(axis=0)) # first normalize from l2-norm
    Y_ = Y.copy()
    Y_ /= np.sqrt((Y_ ** 2).sum(axis=0)) # first normalize from l2-norm
    
    Sc = (X_.T.dot(Y_)) # then derive vector product across columns
    return Sc


def get_SD(img):
    array = sitk.GetArrayFromImage(img)
    std = array.std(axis=0)
    std_image = copyInfo_3DImage(
        sitk.GetImageFromArray(std, isVector=False), img)
    return std_image


def get_motion_R2(timeseries_img, translations,rotations):
    from rabies.analysis_pkg.analysis_math import closed_form
    timeseries_4d = sitk.GetArrayFromImage(timeseries_img)
    timeseries = timeseries_4d.reshape(timeseries_4d.shape[0],-1).astype(float)
    confounds_array = np.concatenate((translations,rotations), axis=1)
    
    X=confounds_array-confounds_array.mean(axis=0)
    Y=timeseries-timeseries.mean(axis=0)
    res = Y-X.dot(closed_form(X,Y))
    R2_spatial = 1-(res.var(axis=0)/Y.var(axis=0))

    R2_img = sitk.GetImageFromArray(R2_spatial.reshape(timeseries_4d.shape[1:]))
    R2_img = copyInfo_3DImage(R2_img, timeseries_img)
    return R2_img


def HMC_derivatives(in_img, in_ref, motcorr_params_file, n_procs=1, get_R2=False):
    import pandas as pd
    from simpleitk_timeseries_motion_correction.apply_transforms import read_transforms_from_csv, framewise_resample_volume

    # the input can be either a nifti file or an SITK image
    if isinstance(in_img, sitk.Image):
        img_preHMC = in_img
    elif os.path.isfile(in_img):
        img_preHMC = sitk.ReadImage(in_img)
    else:
        raise ValueError(f"in_img must be an SITK image or valid file path. Got: {in_img}")
    if isinstance(in_ref, sitk.Image):
        ref_img = in_ref
    elif os.path.isfile(in_ref):
        ref_img = sitk.ReadImage(in_ref)
    else:
        raise ValueError(f"in_ref must be an SITK image or valid file path. Got: {in_ref}")        
    del in_img, in_ref

    # prepare timeseries post-correction
    transforms = read_transforms_from_csv(motcorr_params_file)
    img_postHMC = framewise_resample_volume(
            img_preHMC, 
            ref_img, 
            transforms, 
            interpolation=sitk.sitkLinear, 
            clip_negative=True,
            extrapolator=False,
            max_workers=n_procs,
            )
    
    df = pd.read_csv(motcorr_params_file)
    translations = np.array([df[par] for par in ['TransX', 'TransY', 'TransZ']]).T
    rotations = np.array([df[par] for par in ['EulerX', 'EulerY', 'EulerZ']]).T

    num_volumes = img_preHMC.GetSize()[3]
    ref_img_array = sitk.GetArrayFromImage(ref_img).reshape(-1,1).astype(float)
    timeseries_preHMC = sitk.GetArrayFromImage(img_preHMC).reshape(num_volumes,-1).astype(float)
    timeseries_postHMC = sitk.GetArrayFromImage(img_postHMC).reshape(num_volumes,-1).astype(float)
    
    D_Sc_preHMC = 1-cosine_similarity(timeseries_preHMC.T,ref_img_array)
    D_Sc_postHMC = 1-cosine_similarity(timeseries_postHMC.T,ref_img_array)

    mse_preHMC = np.mean((timeseries_preHMC.T - ref_img_array)**2, axis=0) # taking mean square error
    mse_postHMC = np.mean((timeseries_postHMC.T - ref_img_array)**2, axis=0) # taking mean square error
    del ref_img_array, timeseries_preHMC, timeseries_postHMC
    
    img_preHMC_SD = get_SD(img_preHMC)
    img_postHMC_SD = get_SD(img_postHMC)

    img_preHMC_R2 = get_motion_R2(img_preHMC, translations,rotations) if get_R2 else None
    img_postHMC_R2 = get_motion_R2(img_postHMC, translations,rotations) if get_R2 else None

    derivatives_dict = {'translations':translations, 'rotations':rotations, 
            'D_Sc_preHMC':D_Sc_preHMC, 'D_Sc_postHMC':D_Sc_postHMC,'mse_preHMC':mse_preHMC, 'mse_postHMC':mse_postHMC, 
            'img_preHMC_SD':img_preHMC_SD, 'img_postHMC_SD':img_postHMC_SD, 'img_preHMC_R2':img_preHMC_R2, 'img_postHMC_R2':img_postHMC_R2}

    return img_preHMC, img_postHMC, derivatives_dict


def plot_motion_QC(derivatives_dict, ref_file, plot_R2=False):
    import matplotlib.pyplot as plt
    from rabies.visualization import plot_3d, otsu_scaling
    from matplotlib.gridspec import GridSpec
    nrows = 7
    if plot_R2:
        nrows +=3

    fig = plt.figure(figsize=(8*2,2*nrows), constrained_layout=True)
    gs = GridSpec(nrows, 3, figure=fig)
        
    ###PLOT timecourses
    ax_fused_l = [fig.add_subplot(gs[i, 0:3]) for i in range(4)]
    
    ax=ax_fused_l[0]
    ax.plot(derivatives_dict['translations'])
    ax.set_ylabel('mm', fontsize=15, color='white')
    ax.set_title('Translations', fontsize=20, color='white')
    ax.tick_params(labelsize=15)
    
    ax=ax_fused_l[1]
    ax.plot(derivatives_dict['rotations'])
    ax.set_ylabel('Euler angle (radians)', fontsize=15, color='white')
    ax.set_title('Rotations', fontsize=20, color='white')
    ax.tick_params(labelsize=15)
    
    ax=ax_fused_l[2]
    ax.plot(derivatives_dict['D_Sc_preHMC'])
    ax.plot(derivatives_dict['D_Sc_postHMC'])
    ax.set_ylabel('Cosine distance', fontsize=15, color='white')
    ax.set_title('Difference relative to reference volume', fontsize=20, color='white')
    ax.legend(['Before correction','After correction'], loc="upper right", fontsize=12)
    ax.tick_params(labelsize=15)
    
    ax=ax_fused_l[3]
    ax.plot(derivatives_dict['mse_preHMC'])
    ax.plot(derivatives_dict['mse_postHMC'])
    ax.set_ylabel('Mean square error', fontsize=15, color='white')
    ax.set_title('Difference relative to reference volume', fontsize=20, color='white')
    ax.legend(['Before correction','After correction'], loc="upper right", fontsize=12)
    ax.tick_params(labelsize=15)
    
    ###PLOT REFERENCE
    nrow=4
    ax_3d = [fig.add_subplot(gs[nrow+i, 0]) for i in range(3)]
    ax_3d[0].set_title('Reference image', fontsize=20, color='white')
    cbar_list = plot_3d(ax_3d,otsu_scaling(ref_file),fig=fig,vmin=0,vmax=1,cmap='gray', cbar=False)

    ###PLOT SD
    SD_array = sitk.GetArrayFromImage(derivatives_dict['img_preHMC_SD']).flatten()
    SD_array.sort()
    std_vmax=SD_array[int(len(SD_array)*0.99)]
    
    ax_3d = [fig.add_subplot(gs[nrow+i, 1]) for i in range(3)]
    ax_3d[0].set_title('Temporal standard deviation\n(before correction)', fontsize=20, color='white')
    cbar_list = plot_3d(ax_3d,derivatives_dict['img_preHMC_SD'],fig=fig,vmin=0,vmax=std_vmax,cmap='inferno', cbar=True)
    for cbar in cbar_list:
        cbar.ax.tick_params(labelsize=15)
    
    ax_3d = [fig.add_subplot(gs[nrow+i, 2]) for i in range(3)]
    ax_3d[0].set_title('Temporal standard deviation\n(after correction)', fontsize=20, color='white')
    cbar_list = plot_3d(ax_3d,derivatives_dict['img_postHMC_SD'],fig=fig,vmin=0,vmax=std_vmax,cmap='inferno', cbar=True)
    for cbar in cbar_list:
        cbar.ax.tick_params(labelsize=15)
    
    if plot_R2:
        nrow+=3

        ax_3d = [fig.add_subplot(gs[nrow+i, 0]) for i in range(3)]
        ax_3d[0].set_title('Reference image', fontsize=20, color='white')
        cbar_list = plot_3d(ax_3d,otsu_scaling(ref_file),fig=fig,vmin=0,vmax=1,cmap='gray', cbar=False)
        
        ax_3d = [fig.add_subplot(gs[nrow+i, 1]) for i in range(3)]
        ax_3d[0].set_title('R2 from 6 motion parameters\n(before correction)', fontsize=20, color='white')
        cbar_list = plot_3d(ax_3d,derivatives_dict['img_preHMC_R2'],fig=fig,vmin=0,vmax=1,cmap='inferno', cbar=True)
        for cbar in cbar_list:
            cbar.ax.tick_params(labelsize=15)
        
        ax_3d = [fig.add_subplot(gs[nrow+i, 2]) for i in range(3)]
        ax_3d[0].set_title('R2 from 6 motion parameters\n(after correction)', fontsize=20, color='white')
        cbar_list = plot_3d(ax_3d,derivatives_dict['img_postHMC_R2'],fig=fig,vmin=0,vmax=1,cmap='inferno', cbar=True)
        for cbar in cbar_list:
            cbar.ax.tick_params(labelsize=15)
        
    return fig


import os
import pathlib
from nipype.pipeline import engine as pe
from nipype.interfaces import utility as niu
from nipype.interfaces.io import DataSink
from nipype.interfaces.utility import Function
from rabies.analysis_pkg.diagnosis_pkg.diagnosis_wf import init_diagnosis_wf
from rabies.analysis_pkg.analysis_wf import init_analysis_wf
from rabies.utils import fill_split_dict, get_workflow_dict


def init_main_analysis_wf(preprocess_opts, cr_opts, analysis_opts):

    if analysis_opts.ROI_labels_file is None and analysis_opts.FC_matrix and analysis_opts.ROI_type=='parcellated':
        raise ValueError(
            'No --ROI_labels_file were provided to match the template --anat_template from preprocessing.')

    workflow = pe.Workflow(name='analysis_main_wf')

    conf_output = os.path.abspath(str(analysis_opts.confound_correction_out))

    split_dict, split_name_list, target_list = read_confound_workflow(conf_output, cr_opts)

    number_functional_scans = len(split_name_list)
    if number_functional_scans==0:
        raise ValueError(f"""
            No outputs were founds from the confound correction stage. 
            All scans may have been removed for not meeting the minimum_timepoint threshold 
            when applying --frame_censoring. Outputs will be named empty.nii.gz if this is 
            the case.
            """)

    if analysis_opts.num_ITK_threads=='optimal': # override previous assignment from run_main(): an optimal thread number can be selected based on # of scans
        from nipype import logging
        log = logging.getLogger('nipype.workflow')
        num_ITK_threads=max(int(analysis_opts.local_threads/number_functional_scans),1)
        log.info(f"An optimal number of {num_ITK_threads} ITK threads are allocated for {number_functional_scans} functional scans.")
        os.environ['RABIES_ITK_NUM_THREADS']=str(num_ITK_threads)

    if analysis_opts.group_ica['apply'] and cr_opts.nativespace_analysis and not cr_opts.resample_to_commonspace:
        raise ValueError(
            'No commonspace timeseries are found for running group-ICA. Try applying --resample_to_commonspace to generate these timeseries when using --nativespace_analysis.')

    if analysis_opts.data_diagnosis and cr_opts.nativespace_analysis:
        analysis_opts.resample_to_commonspace = True # resampling to common is necessary for --data_diagnosis report

    # filter inclusion/exclusion lists
    from rabies.utils import filter_scan_inclusion, filter_scan_exclusion
    split_name_list = filter_scan_inclusion(analysis_opts.inclusion_ids, split_name_list)
    split_name_list = filter_scan_exclusion(analysis_opts.exclusion_ids, split_name_list)

    # setting up iterables from the BOLD scan splits
    main_split = pe.Node(niu.IdentityInterface(fields=['split_name']),
                         name="main_split")
    main_split.iterables = [('split_name', split_name_list)]

    # function to read each elements from the dictionary of confound correction outputs
    def read_dict(split_dict, split_name, target_list):
        return [split_dict[split_name][target] for target in target_list]

    # set outputnode from confound correction
    conf_outputnode = pe.Node(Function(input_names=['split_dict', 'split_name', 'target_list'],
                                           output_names=target_list,
                                       function=read_dict),
                              name='conf_outputnode')
    conf_outputnode.inputs.split_dict = split_dict
    conf_outputnode.inputs.target_list = target_list

    # prepare dictionary for seed files
    seed_dict = {}
    for seed_file,seed_name in zip(analysis_opts.seed_list, analysis_opts.seed_name_list):
        seed_dict[seed_name] = seed_file

    if not os.path.isfile(str(analysis_opts.prior_maps)):
        raise ValueError("--prior_maps doesn't exists.")

    load_CR_dict_node = pe.Node(Function(input_names=['maps_dict_file', 'cleaned_bold_file', 'CR_data_dict', 'VE_file', 'STD_file', 'CR_STD_file', 'name_source'],
                                           output_names=[
                                               'CR_dict_file'],
                                       function=load_CR_input_dict),
                              name='load_CR_dict_node')

    # prepare analysis workflow
    analysis_output = os.path.abspath(str(analysis_opts.output_dir))
    analysis_wf = init_analysis_wf(
        opts=analysis_opts)

    '''
    PREPARE DATASINKS
    '''
    if cr_opts.nativespace_analysis:
        # prepare analysis datasink for nativespace computations
        nativespace_analysis_datasink = pe.Node(DataSink(base_directory=analysis_output,
                                            container="nativespace_analysis_datasink"),
                                    name="nativespace_analysis_datasink")
        main_analysis_datasink = nativespace_analysis_datasink

    if analysis_opts.resample_to_commonspace or not cr_opts.nativespace_analysis or analysis_opts.group_ica['apply']:
        # prepare analysis datasink for commonspace computations OR nativespace computations resampled to commonspace
        commonspace_analysis_datasink = pe.Node(DataSink(base_directory=analysis_output,
                                            container="commonspace_analysis_datasink"),
                                    name="commonspace_analysis_datasink")
        if not cr_opts.nativespace_analysis:
            main_analysis_datasink = commonspace_analysis_datasink
        if analysis_opts.group_ica['apply']:
            workflow.connect([
                (analysis_wf, commonspace_analysis_datasink, [
                    ("outputnode.group_ICA_dir", "group_ICA_dir"),
                    ]),
                ])

    data_diagnosis_datasink = pe.Node(DataSink(base_directory=analysis_output,
                                         container="data_diagnosis_datasink"),
                                name="data_diagnosis_datasink")
    
    '''
    CORE WORKFLOW CONNECTIONS
    '''

    workflow.connect([
        (main_split, conf_outputnode, [
            ("split_name", "split_name"),
            ]),
        (conf_outputnode, load_CR_dict_node, [
            ("cleaned_path", "cleaned_bold_file"),
            ("data_dict", "CR_data_dict"),
            ("VE_file_path", "VE_file"),
            ("STD_file_path", "STD_file"),
            ("CR_STD_file_path", "CR_STD_file"),
            ("input_bold", "name_source"),
            ]),
        (load_CR_dict_node, analysis_wf, [
            ("CR_dict_file", "subject_inputnode.CR_dict_file"),
            ]),
        (analysis_wf, main_analysis_datasink, [
            ("outputnode.matrix_data_file", "matrix_data_file"),
            ("outputnode.matrix_fig", "matrix_fig"),
            ("outputnode.corr_map_file", "seed_correlation_maps"),
            ("outputnode.seed_timecourse_csv", "seed_timecourse_csv"),
            ("outputnode.DR_nii_file", "dual_regression_nii"),
            ("outputnode.dual_regression_timecourse_csv", "dual_regression_timecourse_csv"),
            ("outputnode.NPR_prior_timecourse_csv", "NPR_prior_timecourse_csv"),
            ("outputnode.NPR_extra_timecourse_csv", "NPR_extra_timecourse_csv"),
            ("outputnode.NPR_prior_filename", "NPR_prior_filename"),
            ("outputnode.NPR_extra_filename", "NPR_extra_filename"),
            ("outputnode.NPR_optimize_report", "NPR_optimize_report"),
            ]),
        ])
    
    analysis_wf.inputs.group_inputnode.commonspace_template = split_dict[split_name_list[0]]["commonspace_resampled_template"]

    '''
    CONDITIONAL NODES AND CONNECTIONS
    '''

    # if inputs are in commmonspace, then analysis is computed with commonspace files
    # if --data_diagnosis, then commonspace files are needed for plotting results
    if not cr_opts.nativespace_analysis or analysis_opts.data_diagnosis:
        load_maps_dict_common_node = pe.Node(Function(input_names=['mask_file', 'WM_mask_file', 'CSF_mask_file', 'atlas_file', 'anat_ref_file',
                                                            'seed_dict', 'prior_maps', 'transform_list','inverse_list', 'name_source', 'interpolation', 'rabies_data_type'],
                                            output_names=[
                                                'maps_dict_file', 'resampled_seed_filelist', 'resampled_atlas_file'],
                                        function=load_maps_dict),
                                name='load_maps_dict_common_node')
        load_maps_dict_common_node.inputs.atlas_file = analysis_opts.ROI_labels_file
        load_maps_dict_common_node.inputs.seed_dict = seed_dict
        load_maps_dict_common_node.inputs.prior_maps = os.path.abspath(analysis_opts.prior_maps)
        load_maps_dict_common_node.inputs.interpolation = analysis_opts.interpolation_sitk
        load_maps_dict_common_node.inputs.rabies_data_type = analysis_opts.data_type

        split_name = split_name_list[0] # can take the commonspace files from any subject, they are all identical
        load_maps_dict_common_node.inputs.mask_file = split_dict[split_name]["commonspace_mask"]
        load_maps_dict_common_node.inputs.WM_mask_file = split_dict[split_name]["commonspace_WM_mask"]
        load_maps_dict_common_node.inputs.CSF_mask_file = split_dict[split_name]["commonspace_CSF_mask"]
        load_maps_dict_common_node.inputs.anat_ref_file = split_dict[split_name]["commonspace_resampled_template"]
        load_maps_dict_common_node.inputs.transform_list = []
        load_maps_dict_common_node.inputs.inverse_list = []
        load_maps_dict_common_node.inputs.name_source = 'commonspace.nii'

        if len(analysis_opts.seed_list)>0:
            workflow.connect([
                (load_maps_dict_common_node, commonspace_analysis_datasink, [
                    ("resampled_seed_filelist", "commonspace_resampled_seeds"),
                    ]),
                ])
        if analysis_opts.ROI_labels_file is not None:
            workflow.connect([
                (load_maps_dict_common_node, commonspace_analysis_datasink, [
                    ("resampled_atlas_file", "commonspace_resampled_atlas"),
                    ]),
                ])

        # only if inputs are in commonspace then analysis computations are also in commonspace
        if not cr_opts.nativespace_analysis:
            workflow.connect([
                (load_maps_dict_common_node, analysis_wf, [
                    ("maps_dict_file", "subject_inputnode.maps_dict_file"),
                    ]),
                (load_maps_dict_common_node, load_CR_dict_node, [
                    ("maps_dict_file", "maps_dict_file"),
                    ]),
                ])

    # if inputs are in native space, then subject specific maps are loaded
    if cr_opts.nativespace_analysis:
        load_maps_dict_native_node = pe.Node(Function(input_names=['mask_file', 'WM_mask_file', 'CSF_mask_file', 'atlas_file', 'anat_ref_file', 
                                                            'seed_dict', 'prior_maps', 'transform_list','inverse_list', 'name_source', 'interpolation', 'rabies_data_type'],
                                            output_names=[
                                                'maps_dict_file', 'resampled_seed_filelist', 'resampled_atlas_file'],
                                        function=load_maps_dict),
                                name='load_maps_dict_native_node')
        load_maps_dict_native_node.inputs.atlas_file = analysis_opts.ROI_labels_file
        load_maps_dict_native_node.inputs.seed_dict = seed_dict
        load_maps_dict_native_node.inputs.prior_maps = os.path.abspath(analysis_opts.prior_maps)
        load_maps_dict_native_node.inputs.interpolation = analysis_opts.interpolation_sitk
        load_maps_dict_native_node.inputs.rabies_data_type = analysis_opts.data_type

        workflow.connect([
            (conf_outputnode, load_maps_dict_native_node, [
                ("native_brain_mask", "mask_file"),
                ("native_WM_mask", "WM_mask_file"),
                ("native_CSF_mask", "CSF_mask_file"),
                ("native_bold_ref", "anat_ref_file"),
                ("commonspace_to_native_transform_list", "transform_list"),
                ("commonspace_to_native_inverse_list", "inverse_list"),
                ("input_bold", "name_source"),
                ]),
            (load_maps_dict_native_node, analysis_wf, [
                ("maps_dict_file", "subject_inputnode.maps_dict_file"),
                ]),
            (load_maps_dict_native_node, load_CR_dict_node, [
                ("maps_dict_file", "maps_dict_file"),
                ]),
            (conf_outputnode, analysis_wf, [
                ("native_to_commonspace_transform_list", "subject_inputnode.native_to_commonspace_transform_list"),
                ("native_to_commonspace_inverse_list", "subject_inputnode.native_to_commonspace_inverse_list"),
                ]),
            ])
        if len(analysis_opts.seed_list)>0:
            workflow.connect([
                (load_maps_dict_native_node, nativespace_analysis_datasink, [
                    ("resampled_seed_filelist", "nativespace_resampled_seeds"),
                    ]),
                ])
        if analysis_opts.ROI_labels_file is not None:
            workflow.connect([
                (load_maps_dict_native_node, nativespace_analysis_datasink, [
                    ("resampled_atlas_file", "nativespace_resampled_atlas"),
                    ]),
                ])


    if not cr_opts.nativespace_analysis or cr_opts.resample_to_commonspace:
        # this node can only exist if commonspace timeseries were previously generated
        analysis_split_joinnode = pe.JoinNode(niu.IdentityInterface(fields=['file_list', 'mask_file']),
                                            name='analysis_split_joinnode',
                                            joinsource='main_split',
                                            joinfield=['file_list'])

        workflow.connect([
            (conf_outputnode, analysis_split_joinnode, [
                ("commonspace_mask", "mask_file"),
                ]),
            (analysis_split_joinnode, analysis_wf, [
                ("file_list", "group_inputnode.bold_file_list"),
                ("mask_file", "group_inputnode.commonspace_mask"),
                ]),
            ])

        if not cr_opts.nativespace_analysis:
            workflow.connect([
                (conf_outputnode, analysis_split_joinnode, [
                    ("cleaned_path", "file_list"),
                    ]),
                ])
        elif cr_opts.resample_to_commonspace:
            workflow.connect([
                (conf_outputnode, analysis_split_joinnode, [
                    ("cleaned_timeseries_commonspace", "file_list"),
                    ]),
                ])

    # handling the resampling of nativespace computations into commonspace outputs        
    if cr_opts.nativespace_analysis and analysis_opts.resample_to_commonspace:
        workflow.connect([
            (analysis_wf, commonspace_analysis_datasink, [
                ("outputnode.corr_map_file_resampled", "seed_correlation_maps_resampled"),
                ]),
            (analysis_wf, commonspace_analysis_datasink, [
                ("outputnode.DR_nii_file_resampled", "dual_regression_nii_resampled"),
                ]),
            ])

    if analysis_opts.data_diagnosis:

        def prep_analysis_dict(seed_map_files, seed_timecourse_csv, dual_regression_nii, dual_regression_timecourse_csv, NPR_prior_timecourse_csv, NPR_extra_timecourse_csv, NPR_prior_filename, NPR_extra_filename):
            return {'seed_map_files':seed_map_files, 'seed_timecourse_csv':seed_timecourse_csv, 'dual_regression_nii':dual_regression_nii, 'dual_regression_timecourse_csv':dual_regression_timecourse_csv, 
                    'NPR_prior_timecourse_csv':NPR_prior_timecourse_csv, 'NPR_extra_timecourse_csv':NPR_extra_timecourse_csv, 
                    'NPR_prior_filename':NPR_prior_filename, 'NPR_extra_filename':NPR_extra_filename}
        prep_analysis_dict_node = pe.Node(Function(input_names=['seed_map_files', 'seed_timecourse_csv', 'dual_regression_nii', 'dual_regression_timecourse_csv', 'NPR_prior_timecourse_csv', 'NPR_extra_timecourse_csv', 'NPR_prior_filename', 'NPR_extra_filename'],
                                            output_names=[
                                                'analysis_dict'],
                                        function=prep_analysis_dict),
                                name='analysis_prep_analysis_dict')

        diagnosis_wf = init_diagnosis_wf(analysis_opts, cr_opts.nativespace_analysis, preprocess_opts, split_name_list, name="diagnosis_wf")

        workflow.connect([
            (load_CR_dict_node, diagnosis_wf, [
                ("CR_dict_file", "inputnode.CR_dict_file"),
                ]),
            (load_maps_dict_common_node, diagnosis_wf, [
                ("maps_dict_file", "inputnode.common_maps_dict_file"),
                ]),
            (analysis_wf, prep_analysis_dict_node, [
                ("outputnode.dual_regression_timecourse_csv", "dual_regression_timecourse_csv"),
                ]),
            (prep_analysis_dict_node, diagnosis_wf, [
                ("analysis_dict", "inputnode.analysis_dict"),
                ]),
            (diagnosis_wf, data_diagnosis_datasink, [
                ("outputnode.figure_temporal_diagnosis", "figure_temporal_diagnosis"),
                ("outputnode.figure_spatial_diagnosis", "figure_spatial_diagnosis"),
                ("outputnode.analysis_QC", "analysis_QC"),
                ("outputnode.temporal_info_csv", "temporal_info_csv"),
                ("outputnode.spatial_VE_nii", "spatial_VE_nii"),
                ("outputnode.temporal_std_nii", "temporal_std_nii"),
                ("outputnode.CR_prediction_std_nii", "CR_prediction_std_nii"),
                ("outputnode.GS_cov_nii", "GS_cov_nii"),
                ]),
            ])
        if cr_opts.nativespace_analysis:
            workflow.connect([
                (load_maps_dict_native_node, diagnosis_wf, [
                    ("maps_dict_file", "inputnode.sub_maps_dict_file"),
                    ]),
                (analysis_wf, prep_analysis_dict_node, [
                    ("outputnode.DR_nii_file_resampled", "dual_regression_nii"),
                    ]),
                (conf_outputnode, diagnosis_wf, [
                    ("native_to_commonspace_transform_list", "inputnode.native_to_commonspace_transform_list"),
                    ("native_to_commonspace_inverse_list", "inputnode.native_to_commonspace_inverse_list"),
                    ]),
                ])
        else:
            workflow.connect([
                (load_maps_dict_common_node, diagnosis_wf, [
                    ("maps_dict_file", "inputnode.sub_maps_dict_file"),
                    ]),
                (analysis_wf, prep_analysis_dict_node, [
                    ("outputnode.DR_nii_file", "dual_regression_nii"),
                    ]),
                ])

        if (analysis_opts.NPR_temporal_comp>-1) or (analysis_opts.NPR_spatial_comp>-1) or analysis_opts.optimize_NPR['apply']:
            workflow.connect([
                (analysis_wf, prep_analysis_dict_node, [
                    ("outputnode.NPR_prior_timecourse_csv", "NPR_prior_timecourse_csv"),
                    ("outputnode.NPR_extra_timecourse_csv", "NPR_extra_timecourse_csv"),
                    ("outputnode.NPR_prior_filename", "NPR_prior_filename"),
                    ("outputnode.NPR_extra_filename", "NPR_extra_filename"),
                    ]),
                ])
        else:
            prep_analysis_dict_node.inputs.NPR_prior_timecourse_csv = None
            prep_analysis_dict_node.inputs.NPR_extra_timecourse_csv = None
            prep_analysis_dict_node.inputs.NPR_prior_filename = None
            prep_analysis_dict_node.inputs.NPR_extra_filename = None

        if len(analysis_opts.seed_list) > 0:
            workflow.connect([
                (analysis_wf, prep_analysis_dict_node, [
                    ("outputnode.joined_corr_map_file", "seed_map_files"),
                    ("outputnode.joined_seed_timecourse_csv", "seed_timecourse_csv"),
                    ]),
                ])
        else:
            prep_analysis_dict_node.inputs.seed_map_files = []
            prep_analysis_dict_node.inputs.seed_timecourse_csv = None

    return workflow


# this function handles masks/maps that can be either common across subjects in commonspace, or resampled into individual spaces for nativespace analyses
# in the case of commonspace, this node only runs once, hence saving computations
# anat_ref_file is an anatomical 3D image that defines the resampling space inputted to antsApplyTransform
def load_maps_dict(mask_file, WM_mask_file, CSF_mask_file, atlas_file, anat_ref_file,
                   seed_dict, prior_maps, transform_list, inverse_list, name_source, interpolation, rabies_data_type):
    import pickle
    import numpy as np
    import SimpleITK as sitk
    import os
    import pathlib  # Better path manipulation
    from rabies.utils import resample_volumes, antsApplyTransforms
    from rabies.analysis_pkg.utils import compute_edge_mask
    mask_img = sitk.ReadImage(mask_file)
    mask_array = sitk.GetArrayFromImage(mask_img)
    volume_indices = mask_array.astype(bool)
    
    if WM_mask_file is None:
        WM_idx = None
    else:
        WM_idx = sitk.GetArrayFromImage(
            sitk.ReadImage(WM_mask_file))[volume_indices].astype(bool)
    if CSF_mask_file is None:
        CSF_idx = None
    else:
        CSF_idx = sitk.GetArrayFromImage(
            sitk.ReadImage(CSF_mask_file))[volume_indices].astype(bool)

    edge_idx = compute_edge_mask(mask_array, num_edge_voxels=1)[volume_indices]

    resampled_4D_img = resample_volumes(in_img=prior_maps, in_ref=anat_ref_file, transforms_3d_files=transform_list, inverses_3d=inverse_list, 
                                          motcorr_params_file=None, interpolation=interpolation, rabies_data_type=rabies_data_type, clip_negative=False)
    resampled_maps = sitk.GetArrayFromImage(resampled_4D_img)
    prior_map_vectors = resampled_maps[:,volume_indices] # we return the 2D format of map number by voxels

    # resample each seed into the right space and then load the seed as an array
    seed_arr_dict = {}
    seed_name_list = list(seed_dict.keys())
    resampled_seed_filelist = []
    for seed_name in seed_name_list:
        seed_file = seed_dict[seed_name]
        resampled_seed_file = os.path.abspath(f'{seed_name}_resampled.nii.gz')
        antsApplyTransforms(transforms = transform_list, inverses = inverse_list, 
                        input_image = seed_file, ref_image = anat_ref_file, output_filename = resampled_seed_file, interpolation='GenericLabel', rabies_data_type=sitk.sitkInt16, clip_negative=False)
        seed_arr_dict[seed_name] = sitk.GetArrayFromImage(sitk.ReadImage(resampled_seed_file))[volume_indices]
        resampled_seed_filelist.append(resampled_seed_file)

    if len(resampled_seed_filelist)==0:
        resampled_seed_filelist = None

    if atlas_file is None:
        atlas_idx = None
        roi_list = None
        resampled_atlas_file = None
    else:
        resampled_atlas_file = os.path.abspath(f'parcellation_resampled.nii.gz')
        antsApplyTransforms(transforms = transform_list, inverses = inverse_list, 
                        input_image = atlas_file, ref_image = anat_ref_file, output_filename = resampled_atlas_file, interpolation='GenericLabel', rabies_data_type=sitk.sitkInt16, clip_negative=False)
        atlas_idx = sitk.GetArrayFromImage(sitk.ReadImage(resampled_atlas_file))[volume_indices]

        # prepare the list ROI numbers from the atlas for FC matrices
        # get the complete set of original ROI integers 
        atlas_data = sitk.GetArrayFromImage(sitk.ReadImage(str(atlas_file))).astype(int)
        roi_list = []
        for i in range(1, atlas_data.max()+1):
            if np.max(i == atlas_data):  # include integers that do have labelled voxels
                roi_list.append(i)

    maps_dict = {'mask_file':mask_file, 'volume_indices':volume_indices, 'WM_idx':WM_idx, 'CSF_idx':CSF_idx,
                 'atlas_idx':atlas_idx, 'edge_idx':edge_idx, 'anat_ref_file':anat_ref_file,
                 'seed_arr_dict':seed_arr_dict, 'prior_map_vectors':prior_map_vectors, 'roi_list':roi_list}
    
    filename_split = pathlib.Path(name_source).name.rsplit(".nii")
    maps_dict_file = os.path.abspath(f'{filename_split[0]}_maps_dict.pkl')
    with open(maps_dict_file, 'wb') as handle:
        pickle.dump(maps_dict, handle, protocol=pickle.HIGHEST_PROTOCOL)

    return maps_dict_file, resampled_seed_filelist, resampled_atlas_file


# this function loads subject-specific data from confound correction
def load_CR_input_dict(maps_dict_file, cleaned_bold_file, CR_data_dict, VE_file, STD_file, CR_STD_file, name_source):
    import pickle
    import pathlib
    import os
    import numpy as np
    import SimpleITK as sitk

    with open(maps_dict_file, 'rb') as handle:
        maps_dict = pickle.load(handle)

    volume_indices = maps_dict['volume_indices']

    data_img = sitk.ReadImage(cleaned_bold_file)
    data_array = sitk.GetArrayFromImage(data_img)
    num_volumes = data_array.shape[0]
    timeseries = np.zeros([num_volumes, volume_indices.sum()])
    for i in range(num_volumes):
        timeseries[i, :] = (data_array[i, :, :, :])[volume_indices]

    VE_spatial = sitk.GetArrayFromImage(
        sitk.ReadImage(VE_file))[volume_indices]
    temporal_std = sitk.GetArrayFromImage(
        sitk.ReadImage(STD_file))[volume_indices]
    predicted_std = sitk.GetArrayFromImage(
        sitk.ReadImage(CR_STD_file))[volume_indices]

    sub_dict = {'bold_file':cleaned_bold_file, 'name_source':name_source, 'CR_data_dict':CR_data_dict, 
            'timeseries':timeseries, 'VE_spatial':VE_spatial, 'temporal_std':temporal_std,
            'predicted_std':predicted_std}
    
    filename_split = pathlib.Path(name_source).name.rsplit(".nii")
    CR_dict_file = os.path.abspath(f'{filename_split[0]}_data_dict.pkl')
    with open(CR_dict_file, 'wb') as handle:
        pickle.dump(sub_dict, handle, protocol=pickle.HIGHEST_PROTOCOL)

    return CR_dict_file


def read_confound_workflow(conf_output, cr_opts):
    from nipype import logging
    log = logging.getLogger('nipype.workflow')

    conf_workflow_file = f'{conf_output}/rabies_confound_correction_workflow.pkl'

    node_dict = get_workflow_dict(conf_workflow_file)

    preproc_outputnode_name = 'confound_correction_main_wf.preproc_outputnode'
    match_targets = {'input_bold':[preproc_outputnode_name, 'input_bold'],
                    'commonspace_mask':[preproc_outputnode_name, 'commonspace_mask'],
                    'commonspace_WM_mask':[preproc_outputnode_name, 'commonspace_WM_mask'],
                    'commonspace_CSF_mask':[preproc_outputnode_name, 'commonspace_CSF_mask'],
                    'commonspace_vascular_mask':[preproc_outputnode_name, 'commonspace_vascular_mask'],
                    'motion_params_csv':[preproc_outputnode_name, 'motion_params_csv'],
                    'FD_csv':[preproc_outputnode_name, 'FD_csv'],
                    'commonspace_resampled_template':[preproc_outputnode_name, 'commonspace_resampled_template'],
                    'cleaned_path':['confound_correction_main_wf.confound_correction_wf.clean_image', 'cleaned_path'],
                    'data_dict':['confound_correction_main_wf.confound_correction_wf.clean_image', 'data_dict'],
                    'VE_file_path':['confound_correction_main_wf.confound_correction_wf.clean_image', 'VE_file_path'],
                    'STD_file_path':['confound_correction_main_wf.confound_correction_wf.clean_image', 'STD_file_path'],
                    'CR_STD_file_path':['confound_correction_main_wf.confound_correction_wf.clean_image', 'CR_STD_file_path'],
                    'random_CR_STD_file_path':['confound_correction_main_wf.confound_correction_wf.clean_image', 'random_CR_STD_file_path'],
                    'corrected_CR_STD_file_path':['confound_correction_main_wf.confound_correction_wf.clean_image', 'corrected_CR_STD_file_path'],
                    }

    if not cr_opts.nativespace_analysis:
        match_targets.update({
                        'commonspace_bold':[preproc_outputnode_name, 'commonspace_bold'],
                        })
    else:
        match_targets.update({'native_bold':[preproc_outputnode_name, 'native_bold'],
                        'native_bold_ref':[preproc_outputnode_name, 'native_bold_ref'],
                        'native_brain_mask':[preproc_outputnode_name, 'native_brain_mask'],
                        'native_WM_mask':[preproc_outputnode_name, 'native_WM_mask'],
                        'native_CSF_mask':[preproc_outputnode_name, 'native_CSF_mask'],
                        'commonspace_to_native_transform_list':[preproc_outputnode_name, 'commonspace_to_native_transform_list'],
                        'commonspace_to_native_inverse_list':[preproc_outputnode_name, 'commonspace_to_native_inverse_list'],
                        'native_to_commonspace_transform_list':[preproc_outputnode_name, 'native_to_commonspace_transform_list'],
                        'native_to_commonspace_inverse_list':[preproc_outputnode_name, 'native_to_commonspace_inverse_list'],
                        })
        if cr_opts.resample_to_commonspace:
            match_targets.update({'cleaned_timeseries_commonspace':['confound_correction_main_wf.cleaned_bold_to_commonspace', 'resampled_file'],
                        })

    split_dict = {}
    split_name = []
    # preparing a new iterative node where each BOLD scan is a different split
    [unit_bold, output_bold] = match_targets['input_bold']
    bold_dict = node_dict[unit_bold]
    # fill each BOLD scan split with proper affiliated outputs from preprocessing
    fill_split_dict(bold_dict, output_bold, split_name, split_dict, [], node_dict, match_targets)

    target_list = list(match_targets.keys())

    # don't include scans that were removed during confound correction
    corrected_split_name=[]
    remove_list = []
    import pathlib
    for name in split_name:
        filename = pathlib.Path(split_dict[name]['cleaned_path']).name
        if 'empty' in filename:
            remove_list.append(name)
            del split_dict[name]
        else:
            corrected_split_name.append(name)
    split_name = corrected_split_name

    if len(remove_list)>0:
        scan_list_str = ''
        for name in remove_list:
            scan_list_str += f'\n        - {name}'
        log.warning(f"""
        The following scans were not included for analysis as the file was empty: {scan_list_str}
        This is likely due to not meeting the minimum_timepoints threshold from --frame_censoring.
                    """)

    return split_dict, split_name, target_list


def find_split(scan, split_name):
    for split in split_name:
        if split in scan:
            return split
    raise ValueError(f"No previous file name is matching {scan}")


import os
import numpy as np
import SimpleITK as sitk
from rabies.analysis_pkg.analysis_math import closed_form


def tree_list(dirName):
    # Get the list of all files in directory tree at given path
    listOfFiles = list()
    for (dirpath, dirnames, filenames) in os.walk(dirName):
        listOfFiles += [os.path.join(dirpath, file) for file in filenames]
    return listOfFiles


def get_info_list(file_list):
    info_list = []
    for file in file_list:
        basename = os.path.basename(file)
        file_info = basename.split(
            '_run-')[0]+'_run-'+basename.split('_run-')[1][0]
        info_list.append(file_info)

    return info_list


def find_scans(scan_info, bold_files, brain_mask_files, confounds_files, csf_mask_files, FD_files):
    for file in bold_files:
        if scan_info in file:
            bold_file = file
            break
    for file in brain_mask_files:
        if scan_info in file:
            brain_mask_file = file
            break
    for file in confounds_files:
        if scan_info in file:
            confounds_file = file
            break
    for file in csf_mask_files:
        if scan_info in file:
            csf_mask = file
            break
    for file in FD_files:
        if scan_info in file:
            FD_file = file
            break
    return bold_file, brain_mask_file, confounds_file, csf_mask, FD_file


def exec_ICA_AROMA(inFile, mc_file, brain_mask, csf_mask, aroma_out=None, tr='auto', aroma_dim=0, random_seed=1):
    import os
    from rabies.confound_correction_pkg.utils import csv2par
    from rabies.confound_correction_pkg.mod_ICA_AROMA.ICA_AROMA_functions import run_ICA_AROMA
    import pathlib
    filename_split = pathlib.Path(inFile).name.rsplit(".nii")
    if not aroma_out:
        aroma_out = os.getcwd()+'/aroma_out'
    cleaned_file = aroma_out+f'/{filename_split[0]}_aroma.nii.gz'

    if tr=='auto':
        from rabies.utils import get_sitk_header
        header = get_sitk_header(os.path.abspath(inFile))
        tr = float(header.GetSpacing()[3])
    else:
        tr = float(tr)

    success, denoising_applied = run_ICA_AROMA(aroma_out, os.path.abspath(inFile), mc=csv2par(mc_file), TR=float(tr), mask=os.path.abspath(
        brain_mask), mask_csf=os.path.abspath(csf_mask), denType="nonaggr", melDir="", dim=str(aroma_dim), overwrite=True, random_seed=random_seed)
    if not success:
        return None, aroma_out

    # if no denoising was applied, return the input file
    if not denoising_applied:
        return inFile, aroma_out
    else:
        os.rename(aroma_out+'/denoised_func_data_nonaggr.nii.gz', cleaned_file)
        return cleaned_file, aroma_out


def csv2par(in_confounds):
    import pandas as pd
    df = pd.read_csv(in_confounds)
    new_df = pd.DataFrame(
        columns=['mov1', 'mov2', 'mov3', 'rot1', 'rot2', 'rot3'])
    new_df['mov1'] = df['mov1']
    new_df['mov2'] = df['mov2']
    new_df['mov3'] = df['mov3']
    new_df['rot1'] = df['rot1']
    new_df['rot2'] = df['rot2']
    new_df['rot3'] = df['rot3']
    out_confounds = os.path.abspath(
        (os.path.basename(in_confounds).split('.')[0])+('.par'))
    new_df.to_csv(out_confounds, sep='\t', index=False, header=False)
    return out_confounds


def prep_timeseries_interval(timeseries_interval, num_frames):
    if timeseries_interval == 'all':
        raise ValueError(f"'all' is depreciated as an input for --timeseries_interval. Consult rabies confound_correction --help for new syntax. ")
    if ',' in timeseries_interval:
        raise ValueError(f"Use the syntax first-last frame. The syntax with a comma first,last is depreciated.")
    # select the subset of timeseries specified
    split = timeseries_interval.split('-')
    if not len(split)==2:
        raise ValueError(f"--timeseries_interval wasn't split into 2: {split}")
    begin=int(split[0]) # must be an integer
    end=split[1]
    if end=='end':
        end=num_frames # take the total number of time frames
    else:
        end=int(end)

    time_range = range(begin,end)
    return time_range


def compute_FD_censoring(FD_trace, scrubbing_threshold):
    '''
    Scrubbing based on FD: The frames that exceed the given threshold together with 1 back
    and 2 forward frames will be masked out from the data (as in Power et al. 2012)
    '''
    import numpy as np
    cutoff = np.asarray(FD_trace) >= scrubbing_threshold
    mask = np.ones(len(FD_trace)).astype(bool)
    for i in range(len(mask)):
        if cutoff[i]:
            if i==0:
                mask[i:i+3] = 0
            else:
                mask[i-1:i+3] = 0
    return mask


def MSE_relative_to_average(timeseries_4d_arr):
    # timeseries_4d_arr should be a 4d timeseries array before any mean removal,
    # and before any brain mask occured, i.e. we want to include the whole image
    # including non-brain tissue to capture movement and the computation should be
    # driven primarily by image raw constrast
    timeseries_2d_arr = timeseries_4d_arr.reshape(timeseries_4d_arr.shape[0],-1) # reshape to time by voxel array
    # compute an average using trimean
    ref_average_arr = np.quantile(timeseries_2d_arr, (0.2, 0.5, 0.8), axis=0).mean(axis=0)
    mse_trace = np.mean((timeseries_2d_arr - ref_average_arr)**2, axis=1) # taking mean square error
    return mse_trace


def get_DVARS(timeseries):
    # compute the DVARS before denoising
    # the first data point is set to 0
    derivative=np.concatenate((np.zeros((1, timeseries.shape[1])),timeseries[1:,:]-timeseries[:-1,:]))
    DVARS_trace=np.sqrt((derivative**2).mean(axis=1))
    DVARS_trace[0]=None # first index is not sensible
    return DVARS_trace


def outlier_censoring(qc_trace, std_thresh=2.5):
    # this function will iteratively censore each data point that falls std_thresh standard deviation
    # away from the mean, until a distribution is obtained with no such outliers remaining
    mask1=np.zeros(len(qc_trace)).astype(bool)
    mask2=np.ones(len(qc_trace)).astype(bool)
    while ((mask2!=mask1).sum()>0):
        mask1=mask2
        mean=qc_trace[mask1].mean()
        std=qc_trace[mask1].std()
        norm=(qc_trace-mean)/std
        mask2=np.abs(norm)<std_thresh
    return mask2


def temporal_censoring(FD_trace, 
        FD_censoring, FD_threshold, DVARS_trace, DVARS_censoring, 
        mse_trace, MSE_censoring, minimum_timepoint):

    num_frames = len(FD_trace) # FD_trace length must correspond with the timeseries length
    frame_mask = np.ones(num_frames).astype(bool)
    if FD_censoring:
        FD_mask = compute_FD_censoring(FD_trace, FD_threshold)
        frame_mask*=FD_mask
    if DVARS_censoring:
        # usually the first index is nan since the derivative is not sensible for the first timepoint
        nan_idx = np.isnan(DVARS_trace)
        non_nan_idx = nan_idx==False
        DVARS_mask = np.ones(num_frames).astype(bool)
        DVARS_mask[nan_idx] = False # censor all nan values
        DVARS_mask[non_nan_idx] = outlier_censoring(DVARS_trace[non_nan_idx], std_thresh=2.5)
        frame_mask*=DVARS_mask
    if MSE_censoring:
        mse_mask = outlier_censoring(mse_trace, std_thresh=2.5)
        frame_mask*=mse_mask
    if frame_mask.sum()<int(minimum_timepoint):
        from nipype import logging
        log = logging.getLogger('nipype.workflow')
        log.warning(f"CENSORING LEFT LESS THAN {str(minimum_timepoint)} VOLUMES. THIS SCAN WILL BE REMOVED FROM FURTHER PROCESSING.")
        return None

    return frame_mask


def select_motion_regressors(nuisance_regressors,motion_params_csv):
    import pandas as pd
    if ('mot_6' in nuisance_regressors) and ('mot_24' in nuisance_regressors):
        raise ValueError(
            "Can't select both the mot_6 and mot_24 options; must pick one.")

    if isinstance(motion_params_csv, pd.DataFrame):
        motion_params_df = motion_params_csv
    else:
        motion_params_df = pd.read_csv(motion_params_csv)
    keys = motion_params_df.keys()
    conf_keys = []
    # only inputing motion regressors at this stage
    for conf in nuisance_regressors:
        if conf == 'mot_6':
            conf_keys += ['mov1', 'mov2', 'mov3', 'rot1', 'rot2', 'rot3']
        elif conf == 'mot_24':
            conf_keys += [s for s in keys if "rot" in s or "mov" in s]

    return np.asarray(motion_params_df[conf_keys])


def compute_signal_regressors(timeseries, nuisance_regressors, brain_mask_idx, WM_mask_idx, CSF_mask_idx, vascular_mask_idx):
    from nipype import logging
    log = logging.getLogger('nipype.workflow')

    # make sure there are no NaN voxels
    timeseries[np.isnan(timeseries)] = 0

    regressors_array = np.empty([timeseries.shape[0],0])
    for conf,mask_idx in zip(['WM_signal','CSF_signal','vascular_signal','global_signal'],
                                [WM_mask_idx,CSF_mask_idx,vascular_mask_idx,brain_mask_idx]):
        if conf in nuisance_regressors:
            if mask_idx is None:
                raise ValueError(f"Received a None value for mask_idx computing {conf}.")
            elif mask_idx.sum()==0:
                log.warning(f"0 voxels were found for the {conf} mask. No regressors are computed using this mask.")
            else:
                regressor_trace = timeseries.T[mask_idx].mean(axis=0)
                regressors_array = np.append(regressors_array,regressor_trace.reshape(-1,1),axis=1)
    
    if ('aCompCor_5' in nuisance_regressors) or ('aCompCor_percent' in nuisance_regressors):
        if ('aCompCor_5' in nuisance_regressors) and ('aCompCor_percent' in nuisance_regressors):
            raise ValueError(
                "Can't select both the aCompCor_5 and aCompCor_percent options; must pick one.")
        if 'aCompCor_5' in nuisance_regressors:
            method='aCompCor_5'
        elif 'aCompCor_percent' in nuisance_regressors:
            method='aCompCor_percent'
        else:
            raise
        if WM_mask_idx is None or CSF_mask_idx is None:
            raise ValueError(f"WM_mask or CSF_mask are None - cannot compute aCompCor.")

        combined_mask_idx = (WM_mask_idx+CSF_mask_idx) > 0
        if combined_mask_idx.sum()<5:
            log.warning(f"Less than 5 voxels were found within the combined WM and CSF masks - no aCompCor regressors will be computed.")
        else:
            masked_timeseries = timeseries[:,combined_mask_idx]

            from sklearn.decomposition import PCA
            if method == 'aCompCor_percent':
                pca = PCA()
                comp_timeseries = pca.fit_transform(masked_timeseries)
                explained_variance = pca.explained_variance_ratio_
                cum_var = 0
                num_comp = 0
                # evaluate the # of components to explain 50% of the variance
                while(cum_var <= 0.5):
                    cum_var += explained_variance[num_comp]
                    num_comp += 1
                log.info("Extracting "+str(num_comp)+" components for aCompCorr.")
                comp_timeseries = comp_timeseries[:,:num_comp]

            elif method == 'aCompCor_5':
                num_comp = 5
                pca = PCA(n_components=num_comp)
                comp_timeseries = pca.fit_transform(masked_timeseries)

            # double check that a singular matrix is not outputed
            A = comp_timeseries
            is_singular = np.linalg.matrix_rank(A) < min(A.shape)
            if not is_singular:
                regressors_array = np.append(regressors_array,comp_timeseries,axis=1)
            else:
                log.warning(f"aCompCor regressors yielded a singular matrix - these regressors are not included.")

    if regressors_array.shape[1]>0:
        # double check that a singular matrix is not outputed
        A = regressors_array
        is_singular = np.linalg.matrix_rank(A) < min(A.shape)
        if not is_singular:
            return regressors_array
        else:
            log.warning(f"Nuisance regressors yielded a singular matrix - an empty matrix is outputed instead.")
            return np.empty([timeseries.shape[0],0])
    else:
        return regressors_array


def nuisance_regression(timeseries_vol, motion_regressors_array, TR, frame_mask, orig_4d_size, brain_mask, WM_mask, CSF_mask, vascular_mask, nuisance_regressors=[], slicewise_correction_direction='Off', generate_CR_null=False, nipype_log=None):
    volume_idx = sitk.GetArrayFromImage(brain_mask).astype(bool)

    if slicewise_correction_direction=='Off':
        slicewise_correction=False
    else:
        '''
        Prepare the indices for selecting individual slices for slicewise nuisance regression
        '''
        slicewise_correction=True
        if slicewise_correction_direction=='RL':
            slice_direction = 0
        elif slicewise_correction_direction=='AP':
            slice_direction = 1
        elif slicewise_correction_direction=='SI':
            slice_direction = 2
        else:
            raise ValueError(f"slicewise_correction_direction must be one of 'RL', 'AP' or 'SI', got {slicewise_correction_direction} instead.")

    if slicewise_correction:
        # create slicewise masks for managing slicewise operations
        dim_size = orig_4d_size[slice_direction]
        all_slices = list(range(dim_size))
        slice_mask_idx_l = [] # list of the index mask to pick an individual slice
        cleaned_slice_l = [] # the list of slices included post-cleaning
        for slice in all_slices:
            slice_mask_idx = volume_idx.copy().astype(int)
            # multiply by the slice mask values by 2 so that only it can be isolated with ==2
            # here the axis direction is inverted since the array is a conversion from a SITK image
            if slice_direction==0:
                slice_mask_idx[:,:,slice]*=2
            elif slice_direction==1:
                slice_mask_idx[:,slice,:]*=2
            elif slice_direction==2:
                slice_mask_idx[slice,:,:]*=2
            else:
                raise ValueError("Slice direction must be 0, 1 or 2.")
            slice_mask_idx = (slice_mask_idx==2)[volume_idx] # convert to vector that matches timeseries_vol array
            if slice_mask_idx.sum()>0:
                slice_mask_idx_l.append(slice_mask_idx)
                cleaned_slice_l.append(slice)

        # prepare arrays that will be filled
        timeseries_slice_l = [] # each slice will be appended to this list
        VE_total_ratio = 0
        VE_spatial = np.zeros(timeseries_vol.shape[1])
        VE_temporal_l = []

    else:
        # create only a single mask that returns the whole array
        slice_mask_idx_l = [np.ones(timeseries_vol.shape[1]).astype(bool)]
        cleaned_slice_l = None

    num_regressors_ = 0 # set to 0 for now
    for slice_mask_idx in slice_mask_idx_l:
        if slicewise_correction:
            timeseries = timeseries_vol[:,slice_mask_idx]
        else:
            timeseries = timeseries_vol
            del timeseries_vol # minimize memory load

        '''
        #9 - If selected, compute the WM/CSF/vascular signal or aCompCorr and add to list of regressors. This is computed post-AROMA and filtering to 
            minimize re-introduction of previously corrected signal fluctuations.
        '''

        def _mask_idx(mask_img):
            if mask_img is None:
                return None
            return sitk.GetArrayFromImage(mask_img).astype(bool)[volume_idx][slice_mask_idx]
        # indices for each mask are first loaded in vector format that matches the timeseries array
        brain_mask_idx, WM_mask_idx, CSF_mask_idx, vascular_mask_idx = [
            _mask_idx(mask_img) for mask_img in [brain_mask, WM_mask, CSF_mask, vascular_mask]
        ]        
        signal_regressors_array = compute_signal_regressors(timeseries, nuisance_regressors, brain_mask_idx, WM_mask_idx, CSF_mask_idx, vascular_mask_idx)
        # we can now create the full list of regressors
        regressors_array = np.append(motion_regressors_array,signal_regressors_array,axis=1)

        '''
        #10 - Apply confound regression using the selected nuisance regressors.
        '''
        # voxels that have a NaN value are set to 0
        nan_voxels = np.isnan(timeseries).sum(axis=0)>1
        timeseries[:,nan_voxels] = 0

        # estimate the VE from the CR selection, or 6 rigid motion parameters if no CR is applied
        try:
            predicted = regressors_array.dot(closed_form(regressors_array,timeseries))
            residuals = timeseries-predicted
        except np.linalg.LinAlgError:
            if nipype_log:
                nipype_log.warning("SINGULAR MATRIX ERROR DURING CONFOUND REGRESSION. THIS SCAN WILL BE REMOVED FROM FURTHER PROCESSING.")
            return None

        # estimates of variance explained
        VE_total_ratio_slice = 1-(residuals.var()/timeseries.var())
        VE_spatial_slice = 1-(residuals.var(axis=0)/timeseries.var(axis=0))
        VE_temporal_slice = 1-(residuals.var(axis=1)/timeseries.var(axis=1))

        if generate_CR_null:
            # estimate the fit from CR with randomized regressors as in BRIGHT AND MURPHY 2015
            randomized_regressors_array = phase_randomized_regressors(regressors_array, frame_mask, TR=TR)
            predicted_random = randomized_regressors_array.dot(closed_form(randomized_regressors_array,timeseries))
        else:
            predicted_random = np.empty([0,timeseries.shape[1]])

        if len(nuisance_regressors) > 0:
            # if confound regression is applied
            timeseries = residuals

        num_regressors = regressors_array.shape[1] # save the number of regressors applied
        del residuals, regressors_array


        if slicewise_correction:
            timeseries_slice_l.append([timeseries,predicted,predicted_random])
            slice_weight = slice_mask_idx.sum()/volume_idx.sum() # compute what proportion of voxels this slice counts as
            VE_total_ratio += VE_total_ratio_slice*slice_weight # create weighted average from the proportion of voxels
            VE_temporal_l.append(VE_temporal_slice*slice_weight) # create weighted average from the proportion of voxels
            VE_spatial[slice_mask_idx] = VE_spatial_slice
        else:
            timeseries_vol = timeseries
            predicted_vol = predicted
            predicted_random_vol = predicted_random
            VE_total_ratio = VE_total_ratio_slice
            VE_spatial = VE_spatial_slice
            VE_temporal = VE_temporal_slice
        del timeseries, predicted, predicted_random

        if num_regressors>num_regressors_: # keep track of the maximal number of regressors across slices
            num_regressors_ = num_regressors
    
    num_regressors = num_regressors_ # final top # of regressors
    del num_regressors_
    # reconstruct the volumetric array after slicewise processing
    if slicewise_correction:
        num_frames = timeseries_slice_l[0][0].shape[0] # check the finalized number of frames
        timeseries_vol = np.zeros([num_frames, timeseries_vol.shape[1]]) # the number of frames can change, not the number of voxels
        predicted_vol = np.zeros(timeseries_vol.shape)
        if generate_CR_null:
            predicted_random_vol = np.zeros(timeseries_vol.shape)
        else:
            predicted_random_vol = np.empty([0,timeseries_vol.shape[1]])

        for slice_mask_idx,timeseries_slice in zip(slice_mask_idx_l,timeseries_slice_l):
            [timeseries,predicted,predicted_random] = timeseries_slice
            timeseries_vol[:,slice_mask_idx] = timeseries
            predicted_vol[:,slice_mask_idx] = predicted
            if generate_CR_null:
                predicted_random_vol[:,slice_mask_idx] = predicted_random
        del timeseries_slice_l, timeseries_slice, timeseries, predicted, predicted_random

        # compute the sum across slices
        VE_temporal = np.array(VE_temporal_l).sum(axis=0)
        del VE_temporal_l

    return timeseries_vol, predicted_vol, predicted_random_vol, num_regressors, VE_temporal, VE_spatial, VE_total_ratio, cleaned_slice_l


def remove_trend(timeseries, frame_mask, order=1 , time_interval='0-end'):
    '''The timeseries is already censored, we need to create a timeseries that is censored with the same time mask so that it matches'''
    #count number of non-censored timepoints
    num_timepoints = len(frame_mask)

    # we create a centered time axis, to fit an intercept value in the center - Centering improves numerical stability when fitting polynomials (because otherwise high powers of large time indices can explode)
    time=np.linspace(-num_timepoints/2,num_timepoints/2 , num_timepoints)
    time_postcensor = time[frame_mask]
    num_timepoints_postcensor = len(time_postcensor)
    
    # create the design matrix of regressors to remove, essentially just polynomials of the time array, up to the user-specified order (time, time**2, etc)
    X = []
    for k in range(1, order+1):
        X.append(time_postcensor**k) #append time array to list (list of arrays)
    X = np.column_stack(X) if X else np.empty((num_timepoints_postcensor, 0)) # turn list into 2D array; if X is an empty list, will evaluate as false, and the right side ('else') will run, otherwise, the left side, the stack will run

    # Always add intercept as last column (if X is empty, the output will only be a column of intercepts)
    X = np.column_stack([X, np.ones(num_timepoints_postcensor)])

    # prepare the time interval over which to compute the trends; this initial estimate does not take into account censoring
    time_range = prep_timeseries_interval(time_interval, num_frames=num_timepoints)
    first = time_range.start
    last = time_range.stop # .stop actually returns the max value for range(), which is excluded as an index
    # the idx range must be shifted by the number of censored frames before the first index to match the censored timeseries array
    if first>0:
        first -= (frame_mask[:first]==False).sum()
    # similarly for the last index
    last -= (frame_mask[:last]==False).sum()
    time_range = range(first,last) # we re-generated a corrected range

    # fit the trends with linear regression
    Y=timeseries
    W = closed_form(X[time_range, :],Y[time_range, :])

    predicted = X.dot(W) #always subtract the trend over the whole timeseries
    residuals = (Y-predicted) # add back the intercept after
    fitted_intercept = W[-1,:] # predicted intercept

    return residuals, fitted_intercept
    

def lombscargle_mathias(t, x, w):

    '''
    Implementation of Lomb-Scargle periodogram as described in Mathias et al. 2004, 
    and applied in Power et al. 2014
    
    Mathias, A., Grond, F., Guardans, R., Seese, D., Canela, M., & Diebner, H. H. (2004). 
    Algorithms for spectral analysis of irregularly sampled time series. Journal of Statistical Software, 11(1), 1-27.
    
    Power, J. D., Mitra, A., Laumann, T. O., Snyder, A. Z., Schlaggar, B. L., & Petersen, S. E. (2014). 
    Methods to detect, characterize, and remove motion artifact in resting state fMRI. Neuroimage, 84, 320-341.
    
    '''

    if (w == 0).sum()>0:
        raise ZeroDivisionError()

    # Check input sizes
    if t.shape[0] != x.shape[0]:
        raise ValueError("Input arrays do not have the same size.")

    # Create empty array for output periodogram
    cw = np.empty((len(w)))
    sw = np.empty((len(w)))
    theta = np.empty((len(w)))

    w_t = w[:,np.newaxis].dot(t[np.newaxis,:])
    theta = (1/(2*w))*np.arctan( \
        np.sin(2*w_t).sum(axis=1)/ \
        np.cos(2*w_t).sum(axis=1))

        
    wt = (t[:,np.newaxis]-theta[np.newaxis,:])*w
    c=np.cos(wt)
    s=np.sin(wt)
    
    cw = x.T.dot(c)/(c**2).sum(axis=0)
    sw = x.T.dot(s)/(s**2).sum(axis=0)
    
    return cw,sw,theta


def lombscargle_mathias_simulate(t, w, cw, sw,theta):
    # recover simulated timeseries for a given time vector t

    wt = (t[:,np.newaxis]-theta[np.newaxis,:])*w
    c=np.cos(wt)
    s=np.sin(wt)
    
    y = c.dot(cw.T)+s.dot(sw.T)
    return y


def lombscargle_fill(x,time_step,time_mask):
    num_timepoints=len(time_mask)
    time=np.linspace(time_step,num_timepoints*time_step,num_timepoints)

    low_freq = 0.005
    high_freq = 1
    freqs=np.linspace(low_freq,high_freq,1000)
    w = freqs*2*np.pi

    t = time[time_mask]
    cw,sw,theta = lombscargle_mathias(t, x, w)
    
    # simulate over entire time
    t=time
    y = lombscargle_mathias_simulate(t, w, cw, sw,theta)
    
    # standardize according to masked data poaxis=0ints    
    y -= y[time_mask].mean(axis=0)
    y /= y[time_mask].std(axis=0)
    
    # re-scale according to original mean/std
    y *= x.std(axis=0)
    y += x.mean(axis=0)
    
    y_fill = np.zeros(y.shape)
    y_fill[time_mask,:] = x
    y_fill[time_mask==0,:] = y[(time_mask==0)]
    return y_fill


def butterworth(signals, TR, high_pass, low_pass):
    from scipy import signal
    
    critical_freq = []
    if high_pass is not None:
        btype = 'high'
        critical_freq.append(high_pass)
    
    if low_pass is not None:
        btype = 'low'
        critical_freq.append(low_pass)
    
    if len(critical_freq) == 2:
        btype = 'band'
    else:
        critical_freq = critical_freq[0]
    
    order=3
    sos = signal.butter(order, critical_freq, fs=1/TR, btype=btype, output='sos')
    return signal.sosfiltfilt(sos, signals, axis=0)


######### Taken from https://stackoverflow.com/questions/39543002/returning-a-real-valued-phase-scrambled-timeseries
def phaseScrambleTS(ts):
    from scipy.fftpack import fft, ifft
    """Returns a TS: original TS power is preserved; TS phase is shuffled."""
    fs = fft(ts, axis=0)
    pow_fs = np.abs(fs) ** 2.
    phase_fs = np.angle(fs)
    phase_fsr = phase_fs.copy()
    if len(ts) % 2 == 0:
        phase_fsr_lh = phase_fsr[1:int(len(phase_fsr)/2)]
    else:
        phase_fsr_lh = phase_fsr[1:int(len(phase_fsr)/2) + 1]
    np.random.shuffle(phase_fsr_lh)
    if len(ts) % 2 == 0:
        phase_fsr_rh = -phase_fsr_lh[::-1]
        phase_fsr = np.concatenate((np.array((phase_fsr[0],)), phase_fsr_lh,
                                    np.array((phase_fsr[int(len(phase_fsr)/2)],)),
                                    phase_fsr_rh))
    else:
        phase_fsr_rh = -phase_fsr_lh[::-1]
        phase_fsr = np.concatenate((np.array((phase_fsr[0],)), phase_fsr_lh, phase_fsr_rh))
    fsrp = np.sqrt(pow_fs) * (np.cos(phase_fsr) + 1j * np.sin(phase_fsr))
    tsrp = ifft(fsrp, axis=0)
    if not np.allclose(tsrp.imag, np.zeros(tsrp.shape)):
        max_imag = (np.abs(tsrp.imag)).max()
        imag_str = f'\nNOTE: a non-negligible imaginary component was discarded.\n\tMax: {max_imag}'
        print(imag_str)
    return tsrp.real


def phase_randomized_regressors(regressors_array, frame_mask, TR):
    """
    METHOD FROM BRIGHT AND MURPHY 2015
    "The true noise regressors were phase-randomised to create simulated noise regressors with similar frequency 
    distributions to the true noise regressors, but with no relation to the measured head motion or physiology. 
    To achieve this, the frequency spectra of the true noise regressors were obtained using a Fourier transform, 
    and the phase of each frequency in half of the spectrum was randomised and mirrored before performing the 
    inverse Fourier transform (this phase randomization was repeated until the temporal correlation with the true 
    regressor was r b 0.1). The entire set of resulting time-series was then orthogonalised to the complete set 
    of original regressors to make them independent from the true noise."

    """
    from rabies.confound_correction_pkg.utils import lombscargle_fill
    num_conf = regressors_array.shape[1]
    randomized_regressors_array = np.zeros(regressors_array.shape)
    for n in range(num_conf):
        corr=1
        iter=1
        while(corr>0.1):
            x=regressors_array[:,n:n+1]
            # fill missing datapoints to obtain a good reading of frequency spectrum
            y = lombscargle_fill(x,TR,frame_mask)
            # phase randomize
            y_r = phaseScrambleTS(y)
            # re-apply the time mask to the same number of timepoints
            y_m = y_r[frame_mask]
            corr = np.abs(np.corrcoef(x.T,y_m.T)[0,1])
            if iter>100: # set a maximum number of iterations
                from nipype import logging
                log = logging.getLogger('nipype.workflow')
                log.warning("Could not set uncorrelated random regressors!")
                break
            iter += 1
            
        #### impose orthogonality relative to every original regressor      
        y_m -= np.matmul(regressors_array, closed_form(regressors_array, y_m))
        randomized_regressors_array[:,n:n+1] = y_m
    return randomized_regressors_array


def smooth_image(img, fwhm, mask_img):
    from rabies.utils import copyInfo_4DImage, copyInfo_3DImage

    dim = img.GetDimension()
    if not (dim==3 or dim==4):
        raise ValueError("Image must be 3D or 4D")

    # Build the affine matrix from SITK image
    # even thought the affine is in LPS instead of RAS (the nibabel/nilearn convention),
    # the +/- sign of the affine values have no incidence on smooth_array calculations
    affine = sitk_affine_lps(img)
    affine = affine[:3, :3]

    if dim==4:
        array_4d = sitk.GetArrayFromImage(img)
        arr = array_4d.transpose(3,2,1,0) # re-orient the array to match the affine
    elif dim==3:
        array_3d = sitk.GetArrayFromImage(img)
        arr = array_3d.transpose(2,1,0) # re-orient the array to match the affine
    smoothed_arr = _smooth_array(arr, affine, fwhm=fwhm, ensure_finite=True, copy=False)
    del arr
    
    # recover SITK orientation
    if dim==4:
        smoothed_arr = smoothed_arr.transpose(3,2,1,0)
    elif dim==3:
        smoothed_arr = smoothed_arr.transpose(2,1,0)
    
    # smoothing creates leakage around mask boundaries
    # correct for edge effects by dividing by the smoothed mask like FSL https://johnmuschelli.com/fslr/reference/fslsmooth.html
    mask_arr = sitk.GetArrayFromImage(mask_img).transpose(2,1,0) # re-orient the array to match the affine
    smoothed_mask = _smooth_array(mask_arr, affine, fwhm=fwhm, ensure_finite=True, copy=False)
    smoothed_mask = smoothed_mask.transpose(2,1,0) # recover SITK orientation
    mask_arr = mask_arr.transpose(2,1,0) # recover SITK orientation
    
    smoothed_arr /= smoothed_mask # correct for edge effect
    smoothed_arr *= mask_arr # re-apply mask to avoid leakage from smoothing
    smoothed_arr[np.isnan(smoothed_arr)] = 0
    
    # recover SITK img
    if dim==4:
        smoothed_img = copyInfo_4DImage(sitk.GetImageFromArray(
            smoothed_arr, isVector=False), img, img)
    elif dim==3:
        smoothed_img = copyInfo_3DImage(sitk.GetImageFromArray(
            smoothed_arr, isVector=False), img)

    return smoothed_img


def sitk_affine_lps(img):
    '''
    Generate the affine matrix from a SITK image.
    Note that SITK affines follow the LPS coordinate.
    '''
    dim = img.GetDimension()
    direction = np.array(img.GetDirection()).reshape(dim, dim)
    spacing = np.diag(img.GetSpacing())
    origin = np.array(img.GetOrigin())

    affine = np.eye(dim + 1)
    affine[:dim, :dim] = direction @ spacing
    affine[:dim, dim] = origin
    return affine


"""
Reproduction of Nilearn's nilearn.image.smooth_img smoothing function (version 0.7.1)

Original reference:
Nilearn developers. Nilearn: Statistical learning for neuroimaging in Python.
URL: https://nilearn.github.io/
"""
def _smooth_array(arr, affine, fwhm=None, ensure_finite=True, copy=True):
    """Smooth images by applying a Gaussian filter.

    Apply a Gaussian filter along the three first dimensions of `arr`.

    Parameters
    ----------
    arr : :class:`numpy.ndarray`
        4D array, with image number as last dimension. 3D arrays are also
        accepted.

    affine : :class:`numpy.ndarray`
        (4, 4) matrix, giving affine transformation for image. (3, 3) matrices
        are also accepted (only these coefficients are used).
        If `fwhm='fast'`, the affine is not used and can be None.

    fwhm : scalar, :class:`numpy.ndarray`/:obj:`tuple`/:obj:`list`, 'fast' or None, optional
        Smoothing strength, as a full-width at half maximum, in millimeters.
        If a nonzero scalar is given, width is identical in all 3 directions.
        A :class:`numpy.ndarray`, :obj:`tuple`, or :obj:`list` must have 3 elements,
        giving the FWHM along each axis.
        If any of the elements is zero or None, smoothing is not performed
        along that axis.
        If  `fwhm='fast'`, a fast smoothing will be performed with a filter
        [0.2, 1, 0.2] in each direction and a normalisation
        to preserve the local average value.
        If fwhm is None, no filtering is performed (useful when just removal
        of non-finite values is needed).

    ensure_finite : :obj:`bool`, optional
        If True, replace every non-finite values (like NaNs) by zero before
        filtering. Default=True.

    copy : :obj:`bool`, optional
        If True, input array is not modified. True by default: the filtering
        is not performed in-place. Default=True.

    Returns
    -------
    :class:`numpy.ndarray`
        Filtered `arr`.

    Notes
    -----
    This function is most efficient with arr in C order.

    """
    # Here, we have to investigate use cases of fwhm. Particularly, if fwhm=0.
    # See issue #1537
    if isinstance(fwhm, (int, float)) and (fwhm == 0.0):
        import warnings
        warnings.warn("The parameter 'fwhm' for smoothing is specified "
                      "as {0}. Setting it to None "
                      "(no smoothing will be performed)"
                      .format(fwhm))
        fwhm = None
    if arr.dtype.kind == 'i':
        if arr.dtype == np.int64:
            arr = arr.astype(np.float64)
        else:
            arr = arr.astype(np.float32)  # We don't need crazy precision.
    if copy:
        arr = arr.copy()
    if ensure_finite:
        # SPM tends to put NaNs in the data outside the brain
        arr[np.logical_not(np.isfinite(arr))] = 0
    if isinstance(fwhm, str) and (fwhm == 'fast'):
        raise NotImplementedError("The 'fast' option for fwhm is not implemented in rabies confound correction. Please specify a numeric value for fwhm or set to None for no smoothing.")
        arr = _fast_smooth_array(arr)
    elif fwhm is not None:
        from scipy.ndimage import gaussian_filter1d
        fwhm = np.asarray([fwhm]).ravel()
        fwhm = np.asarray([0. if elem is None else elem for elem in fwhm])
        affine = affine[:3, :3]  # Keep only the scale part.
        fwhm_over_sigma_ratio = np.sqrt(8 * np.log(2))  # FWHM to sigma.
        vox_size = np.sqrt(np.sum(affine ** 2, axis=0))
        sigma = fwhm / (fwhm_over_sigma_ratio * vox_size)
        for n, s in enumerate(sigma):
            if s > 0.0:
                gaussian_filter1d(arr, s, output=arr, axis=n)
    return arr